from qtpy.QtCore import QPointF, QRectF, Qt
from qtpy.QtWidgets import QGraphicsWidget, QGraphicsLinearLayout, QSizePolicy

from .NodeItem_CollapseButton import NodeItem_CollapseButton
from .FlowViewProxyWidget import FlowViewProxyWidget
# from .Node import Node
from .NodeItem_Icon import NodeItem_Icon
from .NodeItem_TitleLabel import TitleLabel
from .PortItem import InputPortItem, OutputPortItem


class NodeItemWidget(QGraphicsWidget):
    """The QGraphicsWidget managing all GUI components of a NodeItem in widgets and layouts."""

    def __init__(self, node, node_item):
        super().__init__(parent=node_item)

        self.node = node
        self.node_item = node_item
        self.flow_view = self.node_item.flow_view

        self.body_padding = 6
        self.header_padding = (0, 0, 0, 0)  # theme dependent and hence updated in setup_layout()!

        self.icon = NodeItem_Icon(node, node_item) if node.icon else None
        self.collapse_button = NodeItem_CollapseButton(node, node_item) if node.style == 'normal' else None
        self.title_label = TitleLabel(node, node_item)
        self.main_widget_proxy: FlowViewProxyWidget = None
        if self.node_item.main_widget:
            self.main_widget_proxy = FlowViewProxyWidget(self.flow_view)
            self.main_widget_proxy.setWidget(self.node_item.main_widget)
        self.header_layout: QGraphicsWidget = None
        self.header_widget: QGraphicsWidget = None
        self.body_layout: QGraphicsLinearLayout = None
        self.body_widget: QGraphicsWidget = None
        self.inputs_layout: QGraphicsLinearLayout = None
        self.outputs_layout: QGraphicsLinearLayout = None
        self.setLayout(self.setup_layout())

    def setup_layout(self) -> QGraphicsLinearLayout:

        self.header_padding = self.node_item.session_design.flow_theme.header_padding

        #   main layout
        layout = QGraphicsLinearLayout(Qt.Vertical)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)

        if self.node.style == 'normal':
            self.header_widget = QGraphicsWidget()
            # self.header_widget.setContentsMargins(0, 0, 0, 0)
            self.header_widget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
            self.header_layout = QGraphicsLinearLayout(Qt.Horizontal)
            self.header_layout.setSpacing(5)
            self.header_layout.setContentsMargins(
                *self.header_padding
            )
            if self.icon:
                self.header_layout.addItem(self.icon)
                self.header_layout.setAlignment(self.icon, Qt.AlignVCenter | Qt.AlignLeft)

            self.header_layout.addItem(self.title_label)

            self.header_layout.addItem(self.collapse_button)
            self.header_layout.setAlignment(self.collapse_button, Qt.AlignVCenter | Qt.AlignRight)

            self.header_widget.setLayout(self.header_layout)
            # layout.addItem(self.header_layout)
            layout.addItem(self.header_widget)
            # layout.setAlignment(self.title_label, Qt.AlignTop)
        else:
            self.setZValue(self.title_label.zValue() + 1)

        #   inputs
        self.inputs_layout = QGraphicsLinearLayout(Qt.Vertical)
        self.inputs_layout.setSpacing(2)

        #   outputs
        self.outputs_layout = QGraphicsLinearLayout(Qt.Vertical)
        self.outputs_layout.setSpacing(2)

        #   body
        self.body_widget = QGraphicsWidget()
        # self.body_widget.setContentsMargins(0, 0, 0, 0)
        self.body_layout = QGraphicsLinearLayout(Qt.Horizontal)
        self.body_layout.setContentsMargins(
            self.body_padding,
            self.body_padding,
            self.body_padding,
            self.body_padding
        )

        self.body_layout.setSpacing(4)
        self.body_layout.addItem(self.inputs_layout)
        self.body_layout.setAlignment(self.inputs_layout, Qt.AlignVCenter | Qt.AlignLeft)
        self.body_layout.addStretch()
        self.body_layout.addItem(self.outputs_layout)
        self.body_layout.setAlignment(self.outputs_layout, Qt.AlignVCenter | Qt.AlignRight)

        self.body_widget.setLayout(self.body_layout)

        # layout.addItem(self.body_layout)
        layout.addItem(self.body_widget)

        return layout

    def rebuild_ui(self):
        """Due to some really strange and annoying behaviour of these QGraphicsWidgets, they don't want to shrink
        automatically when content is removed, they just stay large, even with a Minimum SizePolicy. I didn't find a
        way around that yet, so for now I have to recreate the whole layout and make sure the widget uses the smallest
        size possible."""

        # if I don't manually remove the ports from the layouts,
        # they get deleted when setting the widget's layout to None below
        for i in range(len(self.node.inputs)):
            self.inputs_layout.removeAt(0)
        for i in range(len(self.node.outputs)):
            self.outputs_layout.removeAt(0)

        self.setLayout(None)
        self.resize(self.minimumSize())
        self.setLayout(self.setup_layout())

        if self.node_item.collapsed:
            return

        for inp_item in self.node_item.inputs:
            self.add_input_to_layout(inp_item)
        for out_item in self.node_item.outputs:
            self.add_output_to_layout(out_item)

        if self.node_item.main_widget:
            self.add_main_widget_to_layout()

    def update_shape(self):

        self.title_label.update_shape()

        # makes extended node items shrink according to resizing input widgets
        if not self.node_item.initializing:
            self.rebuild_ui()
        # strangely, this only works for small node items without this, not for normal ones

        mw = self.node_item.main_widget
        if mw is not None:  # maybe the main_widget got resized
            # self.main_widget_proxy.setMaximumSize(mw.size())

            # self.main_widget_proxy.setMaximumSize(mw.maximumSize())

            self.main_widget_proxy.setMaximumSize(mw.size())
            self.main_widget_proxy.setMinimumSize(mw.size())

            self.adjustSize()
            self.adjustSize()

        self.body_layout.invalidate()
        self.layout().invalidate()
        self.layout().activate()
        # very essential; repositions everything in case content has changed (inputs/outputs/widget)

        if self.node.style == 'small':

            # making it recompute its true minimumWidth here
            self.adjustSize()

            if self.layout().minimumWidth() < self.title_label.width + 15:
                self.layout().setMinimumWidth(self.title_label.width + 15)
                self.layout().activate()

        w = self.boundingRect().width()
        h = self.boundingRect().height()
        rect = QRectF(QPointF(-w / 2, -h / 2),
                      QPointF(w / 2, h / 2))
        self.setPos(rect.left(), rect.top())

        if not self.node.style == 'normal':
            if self.icon:
                self.icon.setPos(
                    QPointF(-self.icon.boundingRect().width() / 2,
                            -self.icon.boundingRect().height() / 2)
                )
                self.title_label.hide()
            else:
                self.title_label.setPos(
                    QPointF(-self.title_label.boundingRect().width() / 2,
                            -self.title_label.boundingRect().height() / 2)
                )


    def add_main_widget_to_layout(self):
        if self.node.main_widget_pos == 'between ports':
            self.body_layout.insertItem(1, self.main_widget_proxy)
            self.body_layout.insertStretch(2)

        elif self.node.main_widget_pos == 'below ports':
            self.layout().addItem(self.main_widget_proxy)
            self.layout().setAlignment(self.main_widget_proxy, Qt.AlignHCenter)

    def add_input_to_layout(self, inp: InputPortItem):
        if self.inputs_layout.count() > 0:
            self.inputs_layout.addStretch()
        self.inputs_layout.addItem(inp)
        self.inputs_layout.setAlignment(inp, Qt.AlignLeft)

    def insert_input_into_layout(self, index: int, inp: InputPortItem):
        self.inputs_layout.insertItem(index * 2 + 1, inp)  # *2 bcs of the stretches
        self.inputs_layout.setAlignment(inp, Qt.AlignLeft)
        if len(self.node.inputs) > 1:
            self.inputs_layout.insertStretch(index * 2 + 1)  # *2+1 because of the stretches, too

    def remove_input_from_layout(self, inp: InputPortItem):
        self.inputs_layout.removeItem(inp)

        # just a temporary workaround for the issues discussed here:
        # https://forum.qt.io/topic/116268/qgraphicslayout-not-properly-resizing-to-change-of-content
        self.rebuild_ui()

    def add_output_to_layout(self, out: OutputPortItem):
        if self.outputs_layout.count() > 0:
            self.outputs_layout.addStretch()
        self.outputs_layout.addItem(out)
        self.outputs_layout.setAlignment(out, Qt.AlignRight)

    def insert_output_into_layout(self, index: int, out: OutputPortItem):
        self.outputs_layout.insertItem(index * 2 + 1, out)  # *2 because of the stretches
        self.outputs_layout.setAlignment(out, Qt.AlignRight)
        if len(self.node.outputs) > 1:
            self.outputs_layout.insertStretch(index * 2 + 1)  # *2+1 because of the stretches, too

    def remove_output_from_layout(self, out: OutputPortItem):
        self.outputs_layout.removeItem(out)

        # just a temporary workaround for the issues discussed here:
        # https://forum.qt.io/topic/116268/qgraphicslayout-not-properly-resizing-to-change-of-content
        self.rebuild_ui()

    def collapse(self):
        self.body_widget.hide()
        if self.main_widget_proxy:
            self.main_widget_proxy.hide()

    def expand(self):
        self.body_widget.show()
        if self.main_widget_proxy:
            self.main_widget_proxy.show()

    def hide_unconnected_ports(self):
        for inp in self.node_item.inputs:
            if len(inp.port.connections) == 0:
                inp.hide()
        for out in self.node_item.outputs:
            if len(out.port.connections) == 0:
                out.hide()

    def show_unconnected_ports(self):
        for inp in self.node_item.inputs:
            inp.show()
        for out in self.node_item.outputs:
            out.show()



from qtpy.QtWidgets import QVBoxLayout, QWidget, QLineEdit, QScrollArea
from qtpy.QtCore import Qt

from .VarsList_VarWidget import VarsList_VarWidget


class VariablesListWidget(QWidget):
    """Convenience class for a QWidget to easily manage script variables of a script."""

    def __init__(self, vars_manager):
        super(VariablesListWidget, self).__init__()

        self.vars_manager = vars_manager
        self.vars_manager.new_var_created.connect(self.add_new_var)
        self.widgets = []
        self.currently_edited_var = ''
        self.ignore_name_line_edit_signal = False  # because disabling causes firing twice otherwise
        # self.data_type_line_edits = []  # same here

        self.setup_UI()


    def setup_UI(self):
        main_layout = QVBoxLayout()

        self.list_layout = QVBoxLayout()
        self.list_layout.setAlignment(Qt.AlignTop)

        # list scroll area

        self.list_scroll_area = QScrollArea()
        self.list_scroll_area.setWidgetResizable(True)
        self.list_scroll_area.setContentsMargins(0, 0, 0, 0)

        w = QWidget()
        w.setContentsMargins(0, 0, 0, 0)
        w.setLayout(self.list_layout)

        self.list_scroll_area.setWidget(w)

        main_layout.addWidget(self.list_scroll_area)

        # ------------------

        # controls

        self.new_var_name_lineedit = QLineEdit()
        self.new_var_name_lineedit.setPlaceholderText('new var\'s title')
        self.new_var_name_lineedit.returnPressed.connect(self.new_var_LE_return_pressed)

        main_layout.addWidget(self.new_var_name_lineedit)

        # ------------------

        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(main_layout)

        self.recreate_list()


    def recreate_list(self):
        for w in self.widgets:
            w.hide()
            del w

        self.widgets.clear()
        # self.data_type_line_edits.clear()

        for v in self.vars_manager.variables:
            new_widget = VarsList_VarWidget(self, self.vars_manager, v)
            # new_widget.name_LE_editing_finished.connect(self.name_line_edit_editing_finished)
            self.widgets.append(new_widget)

        self.rebuild_list()


    def rebuild_list(self):
        for i in range(self.layout().count()):
            self.list_layout.removeItem(self.layout().itemAt(0))

        for w in self.widgets:
            self.list_layout.addWidget(w)


    def new_var_LE_return_pressed(self):
        name = self.new_var_name_lineedit.text()

        if not self.vars_manager.var_name_valid(name=name):
            return

        self.vars_manager.create_new_var(name=name)


    def add_new_var(self, var):
        self.recreate_list()


    # def name_line_edit_editing_finished(self):
    #     var_widget: VarsList_VarWidget = self.sender()
    #     var_widget.name_line_edit.setEnabled(False)
    #
    #     # search for name issues
    #     new_var_name = var_widget.name_line_edit.text()
    #     for v in self.vars_manager.variables:
    #         if v.name == new_var_name:
    #             var_widget.name_line_edit.setText(self.currently_edited_var.name)
    #             return
    #
    #     var_widget.var.name = new_var_name


    def del_variable(self, var, var_widget):
        self.widgets.remove(var_widget)
        var_widget.setParent(None)
        self.vars_manager.delete_var(var)
        # del self.vars_manager.variables[self.vars_manager.variables.index(var)]
        self.recreate_list()


import logging
import traceback

from .Base import Base

from .NodePort import NodeInput, NodeOutput
from .NodePortBP import NodeInputBP, NodeOutputBP
from .dtypes import DType
from .InfoMsgs import InfoMsgs
from .logging.Logger import Logger
from .tools import serialize, deserialize


class Node(Base):
    """Base class for all node blueprints. Such a blueprint is defined by its class, which is registered
    in the session, and actual node objects are instances of it. The static properties are stored using
    static attributes, which works really well in Python.
    All the main API for programming nodes, i.e. creating subclasses of this class, is defined here."""

    title = ''
    type_ = ''
    # keywords = []  TODO: add this with intelligent search in NodeSelectionWidget
    init_inputs: [NodeInputBP] = []
    init_outputs: [NodeOutputBP] = []
    identifier: str = None  # set by Session if None
    identifier_prefix: str = None


    @classmethod
    def build_identifier(cls):

        full_prefix = (cls.identifier_prefix + '.') if cls.identifier_prefix is not None else ''

        if cls.identifier is None:
            cls.identifier = cls.__name__

        cls.identifier = full_prefix + cls.identifier


    def __init__(self, params):
        Base.__init__(self)

        self.flow, self.session, self.init_config = params
        self.script = self.flow.script
        self.inputs: [NodeInputBP] = []
        self.outputs: [NodeOutputBP] = []
        self.loggers = []
        # self.global_logger, self.errors_logger = self.script.logs_manager.default_loggers.values()

        self.initialized = False

        self.block_init_updates = False
        self.block_updates = False

    def finish_initialization(self):
        """
        Loads all default properties from initial config if it was provided;
        sets up inputs and outputs, enables the logs, loads custom (std) config
        and calls self._initialized()
        """

        if self.init_config:
            self.setup_ports(self.init_config['inputs'], self.init_config['outputs'])

            self.load_custom_config(self.init_config)

        else:
            self.setup_ports()

        self.enable_loggers()
        self._initialized()
        self.initialized = True

        # self.update()

    def load_user_config(self):
        """Loads the component-specific config data that was returned by get_state() previously; prints an exception
        if it fails but doesn't crash because that usually happens when developing nodes"""

        if self.init_config:
            try:
                if type(self.init_config['state data']) == dict:  # backwards compatibility
                    self.set_state(self.init_config['state data'])
                else:
                    self.set_state(deserialize(self.init_config['state data']))
            except Exception as e:
                InfoMsgs.write_err(
                    'Exception while setting data in', self.title, 'node:', e, ' (was this intended?)')


    def setup_ports(self, inputs_config=None, outputs_config=None):

        if not inputs_config and not outputs_config:
            for i in range(len(self.init_inputs)):
                inp = self.init_inputs[i]

                if inp.dtype:
                    self.create_input_dt(label=inp.label, dtype=inp.dtype, add_config=inp.add_config)

                else:
                    self.create_input(inp.label, inp.type_, add_config=self.init_inputs[i].add_config)

            for o in range(len(self.init_outputs)):
                out = self.init_outputs[o]
                self.create_output(out.label, out.type_)

        else:  # when loading saved nodes, the init_inputs and init_outputs are irrelevant
            for inp in inputs_config:
                if 'dtype' in inp:
                    self.create_input_dt(label=inp['label'], dtype=DType.from_str(inp['dtype'])(
                        _load_state=deserialize(inp['dtype state'])), add_config=inp)
                else:
                    self.create_input(label=inp['label'], type_=inp['type'], add_config=inp)

                if 'val' in inp:
                    # this means the input is 'data' and did not have any connections,
                    # so we saved its value which was probably represented by some widget
                    # in the front end which has probably overridden the Node.input() method
                    self.inputs[-1].val = deserialize(inp['val'])

            for out in outputs_config:
                self.create_output(out['label'], out['type'])

    # -----------------------------------------------------------------------------------------------------------------

    def update(self, inp=-1):  # , output_called=-1):
        """'Activates' the node, causing an update_event(); prints an exception if something crashed, but prevents
        the application from crashing in such a case"""

        if self.block_updates:
            InfoMsgs.write('update blocked in', self.title, 'node')
            return

        InfoMsgs.write('update in', self.title, 'node on input', inp)

        try:
            self.update_event(inp)
        except Exception as e:

            InfoMsgs.write_err('EXCEPTION in', self.title, '\n', traceback.format_exc())

    # OVERRIDE
    def update_event(self, inp=-1):
        """Gets called when an input received a signal or some node requested data of an output in exec mode"""

        pass

    def input(self, index: int):
        """Returns the value of a data input.
        If the input is connected, the value of the connected output is used:
        If not, the value of the widget is used."""

        InfoMsgs.write('input called in', self.title, 'Node:', index)
        return self.inputs[index].get_val()

    def exec_output(self, index: int):
        """Executes an exec output, causing activation of all connections"""

        self.outputs[index].exec()

    def set_output_val(self, index, val):
        """Sets the value of a data output causing activation of all connections in data mode"""

        self.outputs[index].set_val(val)

    # OVERRIDE
    def place_event(self):
        """
        place_event() is called once the node object has been fully initialized and placed in the flow.
        When loading content, place_event() is executed *before* the connections are built,
        which is important for nodes that need to update once and, during this process, set output data values,
        to prevent other (later connected) nodes from receiving updates because of that.
        Notice that this method gets executed *every time* the node is added to the flow, which can happen
        multiple times, due to undo/redo operations for example.
        Also note that GUI content is generally not accessible yet from here, for that use view_place_event().
        """

        pass

    # OVERRIDE
    def view_place_event(self):
        """
        view_place_event() is called once all GUI for the node has been created by the frontend.
        Any initial communication to widgets is supposed to happen here, and this method is not called
        when running without gui.
        """

        pass

    # OVERRIDE
    def remove_event(self):
        """Called when the node is removed from the flow; useful for stopping threads and timers etc."""

        pass

    # OVERRIDE
    def _initialized(self):  # not used currently

        """Called once all the node's components (including inputs, outputs) have been initialized"""

        pass

    # OVERRIDE
    def custom_config_data(self) -> dict:
        """Convenience method for saving some std config for all nodes in an editor.
        get_state()/set_state() then stays clean for all specific node subclasses"""

        return {}

    # OVERRIDE
    def load_custom_config(self, data: dict):
        """For loading the data returned by custom_config_data()"""
        pass

    # OVERRIDE
    def get_state(self) -> dict:
        """
        Used to store node-specific custom data that needs to be reloaded when loading a project or pasting copied
        components. All values will be serialized by pickle and base64. The corresponding method for the opposite
        operation is set_state().
        """
        return {}

    # OVERRIDE
    def set_state(self, data: dict):
        """
        Used for reloading node-specific custom data which has been previously returned by get_state()
        """
        pass

    # -----------------------------------------------------------------------------------------------------------------

    # API


    #   LOGGING


    def new_logger(self, title) -> Logger:
        """Requesting a new custom Log"""

        logger = self.script.logs_manager.new_logger(title)
        self.loggers.append(logger)
        return logger

    def disable_loggers(self):
        """Disables custom logs"""

        for logger in self.loggers:
            logger.disable()
            # logger.disabled = True

    def enable_loggers(self):
        """Enables custom logs"""

        for logger in self.loggers:
            logger.enable()
            # logger.enabled = True


    #   PORTS


    def create_input(self, label: str = '', type_: str = 'data', add_config={}, insert: int = None):
        """Creates and adds a new input at index pos"""
        # InfoMsgs.write('create_input called')

        inp = NodeInput(
            node=self,
            type_=type_,
            label_str=label,
            add_config=add_config,
        )

        if insert is not None:
            self.inputs.insert(insert, inp)
        else:
            self.inputs.append(inp)


    def create_input_dt(self, label: str, dtype: DType, add_config={}, insert: int = None):
        """Creates and adds a new data input with a DType"""
        # InfoMsgs.write('create_input called')

        inp = NodeInput(
            node=self,
            type_='data',
            label_str=label,
            dtype=dtype,
            add_config=add_config,
        )

        if insert is not None:
            self.inputs.insert(insert, inp)
        else:
            self.inputs.append(inp)


    def delete_input(self, index: int):
        """Disconnects and removes input"""

        inp: NodeInput = self.inputs[index]

        # break all connections
        for c in inp.connections:
            self.flow.connect_nodes(c.out, inp)

        self.inputs.remove(inp)


    def create_output(self, label: str = '', type_: str = 'data', insert: int = None):
        """Creates and adds a new output"""

        out = NodeOutput(
              node=self,
              type_=type_,
              label_str=label
        )

        if insert is not None:
            self.outputs.insert(insert, out)
        else:
            self.outputs.append(out)


    def delete_output(self, index: int):
        """Disconnects and removes output"""

        out: NodeOutput = self.outputs[index]

        # break all connections
        for c in out.connections:
            self.flow.connect_nodes(out, c.inp)

        self.outputs.remove(out)


    #   VARIABLES


    def get_vars_manager(self):
        """Returns a ref to the script's variables manager"""

        return self.script.vars_manager

    def get_var_val(self, name: str):
        """Gets the value of a script variable"""

        return self.get_vars_manager().get_var_val(name)

    def set_var_val(self, name: str, val):
        """Sets the value of a script variable"""

        return self.get_vars_manager().set_var(name, val)

    def register_var_receiver(self, name: str, method):
        """Registers the node with given method as vars receiver in the script's variables manager to catch
        value changes of any variable with the given name"""

        self.get_vars_manager().register_receiver(self, name, method)

    def unregister_var_receiver(self, name: str, method):
        """Unregisters previously registered node as receiver for value changes of script variables with given name"""

        self.get_vars_manager().unregister_receiver(self, name, method)


    # -----------------------------------------------------------------------------------------------------------------


    def prepare_placement(self):
        """Called from Flow when the nodes gets added"""

        self.enable_loggers()
        self.place_event()


    def prepare_removal(self):
        """Called from Flow when the node gets removed"""

        self.remove_event()
        self.disable_loggers()


    def is_active(self):
        for i in self.inputs:
            if i.type_ == 'exec':
                return True
        for o in self.outputs:
            if o.type_ == 'exec':
                return True
        return False


    def config_data(self) -> dict:
        """
        Returns all metadata of the node in JSON-compatible dict.
        Used to rebuild the Flow when loading a project or pasting components.
        """

        # general attributes
        node_dict = {
            'identifier': self.identifier,
            'state data': serialize(self.get_state()),
            **self.custom_config_data()
        }

        # inputs
        inputs = []
        for i in self.inputs:
            input_dict = i.config_data()
            inputs.append(input_dict)
        node_dict['inputs'] = inputs

        # outputs
        outputs = []
        for o in self.outputs:
            output_dict = o.config_data()
            outputs.append(output_dict)
        node_dict['outputs'] = outputs

        return node_dict



"""A collection of useful functions used by different components."""

import base64
import pickle


def serialize(data) -> str:
    return base64.b64encode(pickle.dumps(data)).decode('ascii')


def deserialize(data):
    return pickle.loads(base64.b64decode(data))





from qtpy.QtCore import QObject

from .FlowView import FlowView


class SessionThreadingBridge(QObject):
    """
    A SessionThreadingBridge instance is a session's interface to the main thread to create the GUI, i.e. FlowViews,
    in the main thread in threaded applications.
    """

    # @staticmethod
    def init_flow_view(self, caller, params):
        view = FlowView(*params)
        caller.tmp_data = view



from qtpy.QtCore import QObject, QPropertyAnimation, Property
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QGraphicsItem


class NodeItemAnimator(QObject):

    # TODO: make this work again...

    def __init__(self, node_item):
        super(NodeItemAnimator, self).__init__()

        self.node_item = node_item
        self.animation_running = False

        self.title_activation_animation = QPropertyAnimation(self, b"p_title_color")
        self.title_activation_animation.setDuration(700)
        self.title_activation_animation.finished.connect(self.finished)
        self.body_activation_animation = QPropertyAnimation(self, b"p_body_color")
        self.body_activation_animation.setDuration(700)

    def start(self):
        self.animation_running = True
        self.title_activation_animation.start()
        self.body_activation_animation.start()

    def stop(self):
        # reset color values. it would just freeze without
        self.title_activation_animation.setCurrentTime(self.title_activation_animation.duration())
        self.body_activation_animation.setCurrentTime(self.body_activation_animation.duration())

        self.title_activation_animation.stop()
        self.body_activation_animation.stop()

    def finished(self):
        self.animation_running = False

    def running(self):
        return self.animation_running

    def reload_values(self):
        self.stop()

        # self.node_item.title_label.update_design()
        self.title_activation_animation.setKeyValueAt(0, self.get_title_color())
        self.title_activation_animation.setKeyValueAt(0.3, self.get_body_color().lighter().lighter())
        self.title_activation_animation.setKeyValueAt(1, self.get_title_color())

        self.body_activation_animation.setKeyValueAt(0, self.get_body_color())
        self.body_activation_animation.setKeyValueAt(0.3, self.get_body_color().lighter())
        self.body_activation_animation.setKeyValueAt(1, self.get_body_color())

    def fading_out(self):
        return self.title_activation_animation.currentTime()/self.title_activation_animation.duration() >= 0.3

    def set_animation_max(self):
        self.title_activation_animation.setCurrentTime(0.3*self.title_activation_animation.duration())
        self.body_activation_animation.setCurrentTime(0.3*self.body_activation_animation.duration())

    def get_body_color(self):
        return self.node_item.color

    def set_body_color(self, val):
        self.node_item.color = val
        QGraphicsItem.update(self.node_item)

    p_body_color = Property(QColor, get_body_color, set_body_color)


    def get_title_color(self):
        return self.node_item.widget.title_label.color

    def set_title_color(self, val):
        self.node_item.widget.title_label.color = val
        # QGraphicsItem.update(self.node_item)

    p_title_color = Property(QColor, get_title_color, set_title_color)


"""
These classes wrap the default ryvencore classes by adding Qt signals to the API calls which are used by the front end.
Same goes for Node and Session which are not part of this file.
"""

from qtpy.QtCore import QObject, Signal

from .ryvencore.logging.LogsManager import LogsManager as RC_LogsManager, Logger as RC_Logger

from .ryvencore.Flow import Flow as RC_Flow
from .ryvencore.Node import Node
from .ryvencore.NodePort import NodePort
from .ryvencore.Connection import Connection, DataConnection as RC_DataConnection

from .ryvencore.script_variables.Variable import Variable
from .ryvencore.script_variables.VarsManager import VarsManager as RC_VarsManager


class LogsManager(RC_LogsManager, QObject):

    new_logger_created = Signal(object)

    def __init__(self, script, create_default_logs=True):
        QObject.__init__(self)
        RC_LogsManager.__init__(self, script=script, create_default_logs=create_default_logs)

    def new_logger(self, title: str):
        logger = RC_LogsManager.new_logger(self, title=title)
        self.new_logger_created.emit(logger)
        return logger


class Logger(RC_Logger, QObject):

    # 'enabled' and 'disabled' are attributes of logging.Logger
    sig_enabled = Signal()
    sig_disabled = Signal()

    def __init__(self, name):
        QObject.__init__(self)
        RC_Logger.__init__(self, name=name)

    def enable(self):
        super().enable()
        self.sig_enabled.emit()

    def disable(self):
        super().disable()
        self.sig_disabled.emit()


class VarsManager(RC_VarsManager, QObject):

    new_var_created = Signal(Variable)
    var_deleted = Signal(Variable)
    var_val_changed = Signal(Variable, object)

    def __init__(self, script, config=None):
        QObject.__init__(self)
        RC_VarsManager.__init__(self, script=script, config=config)

    def create_new_var(self, name: str, val=None) -> Variable:
        """Creates and returns a new script variable and emits new_var_created"""

        v = RC_VarsManager.create_new_var(self, name=name, val=val)
        self.new_var_created.emit(v)
        return v

    def delete_var(self, var: Variable):
        """Deletes a variable and triggers the var_deleted signal."""

        RC_VarsManager.delete_var(self, var=var)
        self.var_deleted.emit(var)

    def set_var(self, name, val) -> bool:
        """Sets the value of an existing script variable.
        Returns true in case of success, false if the var couldn't be found and set."""

        b = RC_VarsManager.set_var(self, name=name, val=val)
        if b:
            var = self.variables[self._get_var_index_from_name(name)]
            self.var_val_changed.emit(var, var.val)
        return b


class Flow(RC_Flow, QObject):

    node_added = Signal(Node)
    node_removed = Signal(Node)
    connection_added = Signal(Connection)
    connection_removed = Signal(Connection)

    connection_request_valid = Signal(bool)
    nodes_created_from_config = Signal(list)
    connections_created_from_config = Signal(list)

    algorithm_mode_changed = Signal(str)

    def __init__(self, session, script):
        QObject.__init__(self)
        RC_Flow.__init__(self, session=session, script=script)

    def add_node(self, node: Node):
        """Stores a node object and emits node_added"""

        RC_Flow.add_node(self, node=node)
        self.node_added.emit(node)

    def remove_node(self, node: Node):
        """Removes a node from internal list without deleting it; emits node_removed"""

        RC_Flow.remove_node(self, node=node)
        self.node_removed.emit(node)

    def add_connection(self, c: Connection):
        """Adds a connection object and emits connection_added"""

        RC_Flow.add_connection(self, c=c)
        self.connection_added.emit(c)

    def remove_connection(self, c: Connection):
        """Removes a connection object without deleting it and emits connection_removed"""

        RC_Flow.remove_connection(self, c=c)
        self.connection_removed.emit(c)

    def check_connection_validity(self, p1: NodePort, p2: NodePort, emit=False) -> bool:
        """Checks whether a considered connect action is legal and emits connection_request_valid if emit=True"""

        valid = RC_Flow.check_connection_validity(self, p1=p1, p2=p2)
        if emit:
            self.connection_request_valid.emit(valid)
        return valid

    def create_nodes_from_config(self, nodes_config: list):
        """Creates Nodes from nodes_config, previously returned by config_data"""

        nodes = RC_Flow.create_nodes_from_config(self, nodes_config=nodes_config)
        self.nodes_created_from_config.emit(nodes)
        return nodes

    def connect_nodes_from_config(self, nodes: [Node], config: list):
        connections = RC_Flow.connect_nodes_from_config(self, nodes=nodes, config=config)
        self.connections_created_from_config.emit(connections)
        return connections

    def set_algorithm_mode(self, mode: str):
        """Sets the algorithm mode of the flow, possible values are 'data' and 'exec'"""

        RC_Flow.set_algorithm_mode(self, mode=mode)
        self.algorithm_mode_changed.emit(self.algorithm_mode())


class DataConnection(RC_DataConnection, QObject):

    activated = Signal(object)

    def __init__(self, params):
        QObject.__init__(self)
        RC_DataConnection.__init__(self, params=params)

    def activate(self, data=None):
        self.activated.emit(data)
        RC_DataConnection.activate(self, data=data)



from qtpy.QtWidgets import QWidget, QHBoxLayout, QLabel, QMenu, QAction
from qtpy.QtGui import QIcon, QDrag
from qtpy.QtCore import QMimeData, Qt, QEvent, QByteArray

import json

from ryvencore_qt.src.GlobalAttributes import Location
from .ListWidget_NameLineEdit import ListWidget_NameLineEdit
from ..tools import shorten
from .EditVal_Dialog import EditVal_Dialog


class VarsList_VarWidget(QWidget):
    """A QWidget representing a single script variable for the VariablesListWidget."""

    def __init__(self, vars_list_widget, vars_manager, var):
        super(VarsList_VarWidget, self).__init__()

        self.vars_manager = vars_manager
        self.var = var
        self.vars_list_widget = vars_list_widget
        self.previous_var_name = ''  # for editing

        self.ignore_name_line_edit_signal = False


        # UI
        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)

        # create icon via label
        variable_icon = QIcon(Location.PACKAGE_PATH+'/resources/pics/variable_picture.png')
        icon_label = QLabel()
        icon_label.setFixedSize(15, 15)
        icon_label.setStyleSheet('border:none;')
        icon_label.setPixmap(variable_icon.pixmap(15, 15))
        main_layout.addWidget(icon_label)

        self.name_line_edit = ListWidget_NameLineEdit(var.name, self)
        self.name_line_edit.setPlaceholderText('name')
        self.name_line_edit.setEnabled(False)
        self.name_line_edit.editingFinished.connect(self.name_line_edit_editing_finished)
        self.name_line_edit.unfocused.connect(self.name_line_edit_editing_finished)

        # name_type_layout = QVBoxLayout()
        # name_type_layout.addWidget(self.name_line_edit)
        main_layout.addWidget(self.name_line_edit)

        self.setLayout(main_layout)



    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            if self.name_line_edit.geometry().contains(event.pos()):
                self.name_line_edit_double_clicked()
                return


    def mousePressEvent(self, event):
        if event.button() == Qt.LeftButton:
            drag = QDrag(self)
            mime_data = QMimeData()
            data_text = self.get_drag_data()
            data = QByteArray(bytes(data_text, 'utf-8'))
            mime_data.setData('text/plain', data)
            drag.setMimeData(mime_data)
            drop_action = drag.exec_()
            return


    def event(self, event):
        if event.type() == QEvent.ToolTip:
            val_str = ''
            try:
                val_str = str(self.var.val)
            except Exception as e:
                val_str = 'couldn\'t stringify value'
            self.setToolTip('val type: '+str(type(self.var.val))+'\nval: '+shorten(val_str, 3000, line_break=True))

        return QWidget.event(self, event)


    def contextMenuEvent(self, event):
        menu: QMenu = QMenu(self)

        delete_action = QAction('delete')
        delete_action.triggered.connect(self.action_delete_triggered)

        edit_value_action = QAction('edit value')
        edit_value_action.triggered.connect(self.action_edit_val_triggered)

        actions = [delete_action, edit_value_action]
        for a in actions:
            menu.addAction(a)

        menu.exec_(event.globalPos())


    def action_delete_triggered(self):
        self.vars_list_widget.del_variable(self.var, self)


    def action_edit_val_triggered(self):
        edit_var_val_dialog = EditVal_Dialog(self, self.var.val)
        accepted = edit_var_val_dialog.exec_()
        if accepted:
            self.vars_manager.set_var(self.var.name, edit_var_val_dialog.get_val())

    def name_line_edit_double_clicked(self):
        self.name_line_edit.setEnabled(True)
        self.name_line_edit.setFocus()
        self.name_line_edit.selectAll()

        # self.vars_list_widget.currently_edited_var = self.var
        self.previous_var_name = self.name_line_edit.text()


    def get_drag_data(self):
        data = {'type': 'variable',
                'name': self.var.name,
                'value': self.var.val}  # value is probably unnecessary
        data_text = json.dumps(data)
        return data_text



    def name_line_edit_editing_finished(self):
        if self.ignore_name_line_edit_signal:
            return

        name = self.name_line_edit.text()

        self.ignore_name_line_edit_signal = True
        # self.name_LE_editing_finished.emit()
        if not self.vars_manager.var_name_valid(name):
            self.name_line_edit.setText(self.previous_var_name)
            return

        # rename var
        self.name_line_edit.setEnabled(False)
        self.var.name = name

        self.ignore_name_line_edit_signal = False



"""The base classes for node custom widgets for nodes."""


class MWB:
    """MainWidgetBase"""

    def __init__(self, params):
        self.node, self.node_item = params

    # OVERRIDE
    def get_state(self) -> dict:
        data = {}
        return data

    # OVERRIDE
    def set_state(self, data: dict):
        pass

    # OVERRIDE
    def remove_event(self):
        pass

    def update_node(self):
        self.node.update()

    def update_node_shape(self):
        self.node_item.update_shape()


class IWB:
    """InputWidgetBase"""

    def __init__(self, params):
        self.input, self.input_item, self.node, self.node_item = params

    # OVERRIDE
    def get_val(self):
        """
        Returns the value that the widget represents for the data input.
        It has to be (pickle) serializable!
        """

        return None

    # OVERRIDE
    def get_state(self) -> dict:
        data = {}
        return data

    # OVERRIDE
    def set_state(self, data: dict):
        pass

    # OVERRIDE
    def remove_event(self):
        pass

    # OVERRIDE
    def val_update_event(self, val):
        pass

    def update_node_input(self, val):
        self.input.update(val)

    def update_node(self):
        self.node.update(self.node.inputs.index(self.input))

    def update_node_shape(self):
        self.node_item.update_shape()



from qtpy.QtCore import QObject, Signal


class FlowSessionThreadInterface(QObject):
    """This class is used to automize communication from GUI components from FlowView to their abstract components,
    currently only implementing the execution of node actions after a right click operation.
    If threading is enabled, the FlowView's instance of this class runs in the abstract thread."""


    def trigger_node_action(self, method, data=None):
        if data is not None:
            method(data)
        else:
            method()



from qtpy.QtWidgets import QWidget, QHBoxLayout, QLabel, QMenu, QAction
from qtpy.QtGui import QIcon, QImage
from qtpy.QtCore import Qt, QEvent

import json

from ryvencore_qt.src.GlobalAttributes import Location
from .ListWidget_NameLineEdit import ListWidget_NameLineEdit
from ..ryvencore.MacroScript import MacroScript


class ScriptsList_ScriptWidget(QWidget):
    """A QWidget representing a single Script for the ScriptsListWidget."""

    def __init__(self, scripts_list_widget, session, script):
        super(ScriptsList_ScriptWidget, self).__init__()

        self.session = session
        self.script = script
        self.flow_view = self.session.flow_views[script]
        self.scripts_list_widget = scripts_list_widget
        self.previous_script_title = ''
        self._thumbnail_source = ''
        self.ignore_title_line_edit_signal = False


        # UI
        main_layout = QHBoxLayout()
        main_layout.setContentsMargins(0, 0, 0, 0)

        # create icon via label
        if isinstance(script, MacroScript):
            script_icon = QIcon(Location.PACKAGE_PATH+'/resources/pics/macro_script_picture.png')
        else:
            script_icon = QIcon(Location.PACKAGE_PATH+'/resources/pics/script_picture.png')
        icon_label = QLabel()
        icon_label.setFixedSize(20, 20)
        icon_label.setStyleSheet('border:none;')
        icon_label.setPixmap(script_icon.pixmap(20, 20))
        main_layout.addWidget(icon_label)

        self.title_line_edit = ListWidget_NameLineEdit(script.title, self)
        self.title_line_edit.setPlaceholderText('title')
        self.title_line_edit.setEnabled(False)
        self.title_line_edit.editingFinished.connect(self.title_line_edit_editing_finished)
        self.title_line_edit.unfocused.connect(self.title_line_edit_editing_finished)

        # name_type_layout = QVBoxLayout()
        # name_type_layout.addWidget(self.title_line_edit)
        main_layout.addWidget(self.title_line_edit)

        self.setLayout(main_layout)



    def mouseDoubleClickEvent(self, event):
        if event.button() == Qt.LeftButton:
            if self.title_line_edit.geometry().contains(event.pos()):
                self.title_line_edit_double_clicked()
                return


    def event(self, event):
        if event.type() == QEvent.ToolTip:
            img: QImage = self.flow_view.get_viewport_img()
            self._thumbnail_source = Location.PACKAGE_PATH+'/temp/script_' + self.script.title + '_thumbnail.png'
            img.save(self._thumbnail_source)
            self.setToolTip('<img height=100 src="' + self._thumbnail_source + '"/>')

        return QWidget.event(self, event)


    def contextMenuEvent(self, event):
        menu: QMenu = QMenu(self)

        delete_action = QAction('delete')
        delete_action.triggered.connect(self.action_delete_triggered)

        actions = [delete_action]
        for a in actions:
            menu.addAction(a)

        menu.exec_(event.globalPos())


    def action_delete_triggered(self):
        self.scripts_list_widget.del_script(self.script, self)


    def title_line_edit_double_clicked(self):
        self.title_line_edit.setEnabled(True)
        self.title_line_edit.setFocus()
        self.title_line_edit.selectAll()

        # self.scripts_list_widget.currently_edited_script = self.script
        self.previous_script_title = self.title_line_edit.text()


    def get_drag_data(self):
        """not used so far..."""
        data = {'type': 'script',
                'title': self.script.title}
        data_text = json.dumps(data)
        return data_text


    def title_line_edit_editing_finished(self):
        if self.ignore_title_line_edit_signal:
            return

        title = self.title_line_edit.text()

        self.ignore_title_line_edit_signal = True
        # self.title_LE_editing_finished.emit()
        if not self.session.script_title_valid(title):
            self.title_line_edit.setText(self.previous_script_title)
            return

        # rename script
        self.title_line_edit.setEnabled(False)
        self.script.title = title
        self.session.rename_script(script=self.script, title=title)

        self.ignore_title_line_edit_signal = False



"""This module should actually rather be part of Ryven, but the retain mechanism is needed in std input widgets,
so I decided to keep it in the core as a feature for now. Source code editing could become a common feature for
ryvencore too."""


def retain(foo):
    """This method is very important for live source code manipulation. It ensures, that every reference to an object's
    method is going to change when the method has been edited (and therefore been overridden). If a reference to a
    method is made by providing the method object without calling retain(self.mymethod) or M(self.mymethod), changing
    this method in Ryven will not result in different behavior when the previously created reference is being called."""

    return lambda *args, **kwargs: getattr(foo.__self__, foo.__name__)(*args, **kwargs)


class M:
    def __init__(self, method):
        self.method_name = method.__name__
        self.method = retain(method)

    def __call__(self, *args, **kwargs):
        self.method(*args, **kwargs)



from qtpy.QtCore import QSize, QRectF, QPointF, QSizeF, Qt
from qtpy.QtWidgets import QGraphicsWidget, QGraphicsLayoutItem

from ryvencore_qt.src.GlobalAttributes import Location
from .tools import change_svg_color


class NodeItem_CollapseButton(QGraphicsWidget):
    def __init__(self, node, node_item):
        super().__init__(parent=node_item)

        self.node = node
        self.node_item = node_item

        self.size = QSize(14, 7)

        self.setGraphicsItem(self)
        self.setCursor(Qt.PointingHandCursor)


        self.collapse_pixmap = change_svg_color(Location.PACKAGE_PATH+'/resources/node_collapse_icon.svg',
                                                self.node.color)
        self.expand_pixmap = change_svg_color(Location.PACKAGE_PATH+'/resources/node_expand_icon.svg',
                                              self.node.color)


    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.size)

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())

    def sizeHint(self, which, constraint=...):
        return QSizeF(self.size.width(), self.size.height())

    def mousePressEvent(self, event):
        event.accept()  # make sure the event doesn't get passed on
        self.node_item.flow_view.mouse_event_taken = True

        if self.node_item.collapsed:
            self.node_item.expand()
        else:
            self.node_item.collapse()

    # def hoverEnterEvent(self, event):

    def paint(self, painter, option, widget=None):

        # doesn't work: ...
        # painter.setRenderHint(QPainter.Antialiasing, True)
        # painter.setRenderHint(QPainter.HighQualityAntialiasing, True)
        # painter.setRenderHint(QPainter.SmoothPixmapTransform, True)

        if not self.node_item.hovered:
            return

        if self.node_item.collapsed:
            pixmap = self.expand_pixmap
        else:
            pixmap = self.collapse_pixmap

        painter.drawPixmap(
            0, 0,
            self.size.width(), self.size.height(),
            pixmap
        )



from .Base import Base


from .Script import Script
from .MacroScript import MacroScript
from .InfoMsgs import InfoMsgs
from .RC import CLASSES
from .Node import Node


class Session(Base):
    """The Session is the top level interface to an editor, it represents a project and manages all project-wide
    components, i.e. Scripts, MacroScripts and Node blueprints."""

    def __init__(
            self,
            gui: bool = False,
    ):
        Base.__init__(self)

        # register classes
        self.register_default_classes()

        # initialize node classes for MacroScripts with correct Node base
        MacroScript.build_node_classes()

        # ATTRIBUTES
        self.scripts: [Script] = []
        self.macro_scripts: [MacroScript] = []
        self.nodes = []  # list of node CLASSES
        self.invisible_nodes = [MacroScript.MacroInputNode, MacroScript.MacroOutputNode]  # might change that system in the future
        self.gui: bool = gui


    def register_default_classes(self):
        """
        Registers the default ryvencore-internal implementations of all exposed classes that COULD have been
        extended by the frontend, in which case we don't set the class.
        So, if the frontend extended a class like Node, it will have set CLASSES['node base'] to its own subclass,
        so we leave it as it is then. It is assumed that the frontend's extensions work properly and don't modify
        the functionality to the backend.
        """

        if not CLASSES['node base']:
            CLASSES['node base'] = Node

        if not CLASSES['data conn']:
            from .Connection import DataConnection
            CLASSES['data conn'] = DataConnection

        if not CLASSES['exec conn']:
            from .Connection import ExecConnection
            CLASSES['exec conn'] = ExecConnection

        if not CLASSES['logs manager']:
            from .logging.LogsManager import LogsManager
            CLASSES['logs manager'] = LogsManager

        if not CLASSES['logger']:
            from .logging.Logger import Logger
            CLASSES['logger'] = Logger

        if not CLASSES['vars manager']:
            from .script_variables.VarsManager import VarsManager
            CLASSES['vars manager'] = VarsManager

        if not CLASSES['flow']:
            from .Flow import Flow
            CLASSES['flow'] = Flow


    def register_nodes(self, node_classes: list):
        """Registers a list of Nodes which you then can access in all scripts"""

        for n in node_classes:
            self.register_node(n)


    def register_node(self, node_class):
        """Registers a Node which then can be accessed in all scripts"""

        # build node class identifier
        node_class.build_identifier()

        self.nodes.append(node_class)


    def unregister_node(self, node_class):
        """Unregisters a Node which will then be removed from the available list.
        Existing instances won't be affected."""

        self.nodes.remove(node_class)


    def create_script(self, title: str = None, create_default_logs=True,
                      config: dict = None) -> Script:

        """Creates and returns a new script.
        If a config is provided the title parameter will be ignored."""

        script = Script(
            session=self, title=title, create_default_logs=create_default_logs,
            config_data=config
        )

        self.scripts.append(script)
        script.load_flow()

        return script


    def create_macro(self, title: str = None, create_default_logs=True,
                     config: dict = None) -> MacroScript:

        """
        Creates and returns a new macro script.
        If a config is provided the title parameter will be ignored and the script will not be initialized,
        which you need to do manually after you made sure that the config doesnt contain other macro nodes
        that have not been loaded yet.
        """

        macro_script = MacroScript(
            session=self, title=title, create_default_logs=create_default_logs,
            config_data=config
        )

        self.scripts.append(macro_script)
        self.macro_scripts.append(macro_script)

        # if config is provided, the script's flow contains content that might include
        # macro nodes that are not loaded yet, so initialization is triggered manually from outside then
        if not config:
            macro_script.load_flow()

        return macro_script


    # def all_scripts(self) -> list:
    #     """Returns a list containing all scripts and macro scripts"""
    #     return self.macro_scripts + self.scripts


    def rename_script(self, script: Script, title: str) -> bool:
        """Renames an existing script"""

        if self.script_title_valid(title):
            script.title = title
            return True
        else:
            return False


    def script_title_valid(self, title: str) -> bool:
        """Checks whether a considered title for a new script is valid (unique) or not"""

        if len(title) == 0:
            return False
        for s in self.scripts:  # all_scripts():
            if s.title == title:
                return False

        return True


    def delete_script(self, script: Script):
        """Deletes an existing script. If the script is a macro script, the macro node is unregistered."""

        if isinstance(script, MacroScript):
            self.unregister_node(script.macro_node_class)
            self.macro_scripts.remove(script)
            self.scripts.remove(script)
        else:
            self.scripts.remove(script)


    def info_messenger(self):
        """Returns a reference to InfoMsgs to print info data"""
        return InfoMsgs


    def load(self, project: dict) -> [Script]:
        """Loads a project and raises an exception if required nodes are missing"""

        if 'scripts' not in project and 'macro scripts' not in project:
            raise Exception('not a valid project dict')

        new_macro_scripts = []
        if 'macro scripts' in project:
            for msc in project['macro scripts']:
                new_macro_scripts.append(self.create_macro(config=msc))

            # now all macro nodes have been registered, so we can initialize the scripts

            for ms in new_macro_scripts:
                ms.load_flow()

        new_scripts = []
        for sc in project['scripts']:
            new_scripts.append(self.create_script(config=sc))

        return new_scripts + new_macro_scripts


    def serialize(self) -> dict:
        """Returns the project as JSON compatible dict to be saved and loaded again using load()"""

        data = {}

        macro_scripts_list = []
        for m_script in self.macro_scripts:
            macro_scripts_list.append(m_script.serialize())
        data['macro scripts'] = macro_scripts_list

        scripts_list = []
        for script in set(self.scripts) - set(self.macro_scripts):  # exclude macro scripts
            scripts_list.append(script.serialize())
        data['scripts'] = scripts_list

        return data


    def all_node_objects(self) -> list:
        """Returns a list containing all Node objects used in any flow which is useful for advanced project analysis"""

        nodes = []
        for s in self.scripts:  # all_scripts():
            for n in s.flow.nodes:
                nodes.append(n)
        return nodes


    # def save_as_project(self, fpath: str):
    #     """Convenience method for directly saving the the all content as project to a file"""
    #     pass



from .InfoMsgs import InfoMsgs
from .RC import *
from .Session import Session
from .logging import *
from .Node import Node
from .NodePortBP import NodeInputBP, NodeOutputBP
from .Connection import DataConnection, ExecConnection
from .tools import serialize, deserialize



from qtpy.QtWidgets import QLineEdit
from qtpy.QtCore import Signal


class ListWidget_NameLineEdit(QLineEdit):

    unfocused = Signal()

    def __init__(self, text, parent):
        super(ListWidget_NameLineEdit, self).__init__(text, parent)


    def focusOutEvent(self, arg__1):
        self.unfocused.emit()
        QLineEdit.focusOutEvent(self, arg__1)


from qtpy.QtCore import Qt, QPointF, QRectF
from qtpy.QtGui import QColor, QPainter, QBrush, QRadialGradient, QLinearGradient, QPen, QPainterPath, QFont
from qtpy.QtWidgets import QStyle, QStyleOption

from .tools import pythagoras


class FlowTheme:
    
    # STATIC ATTRIBUTES
    name = ''
    type_ = 'dark'

    node_selection_stylesheet__base = '''
PlaceNodeWidget {
    background: transparent;
}
QScrollArea {
    border-radius: 8px;
}
NodeWidget {
    
}
'''
    node_selection_stylesheet = ''

    header_padding = (10, 2, 10, 2)  # (left, top, right, botton)
    
    exec_conn_color = QColor('#ffffff')
    exec_conn_width = 1.5
    exec_conn_pen_style = Qt.SolidLine
    
    data_conn_color = QColor('#ffffff')
    data_conn_width = 1.5
    data_conn_pen_style = Qt.DashLine
    
    flow_background_brush = QBrush(QColor('#333333'))
    flow_background_grid = None
    flow_highlight_pen_color = QColor('#245d75')

    node_item_shadow_color = QColor('#2b2b2b')

    EXPORT = []


    def __init__(self):
        pass

    def load(self, data: dict):
        if data and self.name in data.keys():
            imported = {}
            for k, v in data[self.name].items():
                if v != 'default':
                    imported[k] = v
            self._load(imported)

    def _load(self, imported: dict):
        for k, v in imported.items():

            if k == 'exec connection color':
                self.exec_conn_color = self.hex_to_col(v)
            elif k == 'exec connection width':
                self.exec_conn_width = v
            elif k == 'exec connection pen style':
                self.exec_conn_pen_style = self._parse_pen_style(v)

            elif k == 'data connection color':
                self.data_conn_color = self.hex_to_col(v)
            elif k == 'data connection width':
                self.data_conn_width = v
            elif k == 'data connection pen style':
                self.data_conn_pen_style = self._parse_pen_style(v)

            elif k == 'flow background color':
                self.flow_background_brush.setColor(self.hex_to_col(v))

    def build_node_selection_stylesheet(self):
        return self.node_selection_stylesheet__base + '\n' + self.node_selection_stylesheet

    def paint_NI_title_label(self, node, painter: QPainter, option: QStyleOption, hovering: bool, node_style: str, 
                             node_title: str, node_color: QColor, node_item_bounding_rect):
        pass

    def paint_PI_label(self, node, painter: QPainter, option: QStyleOption, type_: str, connected: bool, label_str: str, 
                       node_color: QColor, bounding_rect: QRectF):
        pass

    def paint_PI(self, node, painter: QPainter, option: QStyleOption, node_color: QColor, type_: str, connected: bool,
                 padding, w, h):
        pass

    def paint_NI(self, node, node_style: str,
                 painter: QPainter, option: QStyleOption,
                 color: QColor, w, h, bounding_rect, title_rect):

        painter.setRenderHint(QPainter.Antialiasing)

        if node_style == 'normal':
            self.draw_NI_normal(node, painter, color, w, h, bounding_rect, title_rect)
        elif node_style == 'small':
            self.draw_NI_small(node, painter, color, w, h, bounding_rect)

    def draw_NI_normal(self, node, painter: QPainter, c: QColor, w, h, bounding_rect, title_rect):
        pass

    def draw_NI_small(self, node, painter: QPainter, c: QColor, w, h, bounding_rect, background_color=None):
        pass





    @staticmethod
    def paint_NI_title_label_default(painter: QPainter, node_style: str, title: str, color: QColor, pen_w: float,
                                     font: QFont, node_item_bounding_rect):
        pen = QPen(color)
        pen.setWidth(pen_w)

        painter.setPen(pen)
        painter.setFont(font)

        text_rect = node_item_bounding_rect
        text_rect.setTop(text_rect.top())

        if node_style == 'normal':
            painter.drawText(text_rect, Qt.AlignLeft | Qt.AlignVCenter, title)
        elif node_style == 'small':
            painter.drawText(text_rect, Qt.AlignTop | Qt.AlignHCenter, title)

    @staticmethod
    def paint_PI_label_default(painter: QPainter, label_str: str, color: QColor, font: QFont, bounding_rect: QRectF):
        painter.setBrush(Qt.NoBrush)
        pen = QPen(color)
        painter.setPen(pen)
        painter.setFont(font)
        painter.drawText(bounding_rect, Qt.AlignCenter, label_str)

    @staticmethod
    def get_header_rect(node_width, node_height, title_rect):
        header_height = 1.0 * title_rect.height()  # 35 * (self.parent_node.title.count('\n')+1)

        header_rect = QRectF()
        header_rect.setTopLeft(QPointF(-node_width / 2, -node_height / 2))
        header_rect.setWidth(node_width)
        header_rect.setHeight(header_height)
        return header_rect
    
    @staticmethod
    def interpolate_color(c1, c2, val):
        r1 = c1.red()
        g1 = c1.green()
        b1 = c2.blue()
        a1 = c1.alpha()

        r2 = c2.red()
        g2 = c2.green()
        b2 = c2.blue()
        a2 = c2.alpha()

        r = (r2 - r1) * val + r1
        g = (g2 - g1) * val + g1
        b = (b2 - b1) * val + b1
        a = (a2 - a1) * val + a1

        return QColor(r, g, b, a)

    @staticmethod
    def hex_to_col(hex_str: str) -> QColor:
        """Converts a hex value in format '#xxxxxx[xx]' to QColor using alpha value if [xx] is used."""

        h = hex_str.lstrip('#')

        if len(h) == 6:
            r, g, b = tuple(
                int(h[i:i + 2], 16) for i in (0, 2, 4)
            )
            return QColor(r, g, b)
        elif len(h) == 8:
            r, g, b, a = tuple(
                int(h[i:i + 2], 16) for i in (0, 2, 4, 6)
            )
            return QColor(r, g, b, a)

        return None

    @staticmethod
    def _parse_pen_style(s: str):
        if s == 'solid line':
            return Qt.SolidLine
        elif s == 'dash line':
            return Qt.DashLine
        elif s == 'dash dot line':
            return Qt.DashDotLine
        elif s == 'dash dot dot line':
            return Qt.DashDotDotLine
        elif s == 'dot line':
            return Qt.DotLine


class FlowTheme_Toy(FlowTheme):


    name = 'Toy'

    node_selection_stylesheet = ''

    header_padding = (12, 5, 10, 2)

    exec_conn_color = QColor(188, 187, 242)
    exec_conn_width = 5
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(188, 187, 242)
    data_conn_width = 5
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#333333'))


    def paint_NI_title_label(self, node, painter, option, hovering, node_style, node_title, node_color, 
                             node_item_bounding_rect):
        
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=QColor(30, 43, 48) if not hovering else node_color.lighter(),
                pen_w=2 if hovering else 1.5,
                font=QFont('Poppins', 15),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=QColor(30, 43, 48) if not hovering else node_color.lighter(),
                pen_w=1.5,
                font=QFont('K2D', 20, QFont.Bold, True),
                node_item_bounding_rect=node_item_bounding_rect
            )
    
    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = QColor('#FFFFFF')
        self.paint_PI_label_default(painter, label_str, c, QFont("Source Code Pro", 10, QFont.Bold), bounding_rect)
    
    def paint_PI(self, node, painter, option, node_color, type_, connected, padding, w, h):

        color = QColor('#2E688C') if type_ == 'data' else QColor('#3880ad')
        if option.state & QStyle.State_MouseOver:
            color = color.lighter()

        brush = QBrush(QColor(color))
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)

        painter.drawEllipse(QRectF(padding, padding, w, h))


    def draw_NI_normal(self, node, painter: QPainter, c: QColor, w, h, bounding_rect, title_rect):

        # main rect
        body_gradient = QRadialGradient(bounding_rect.topLeft(), pythagoras(h, w))
        body_gradient.setColorAt(0, QColor(c.red() / 10 + 100, c.green() / 10 + 100, c.blue() / 10 + 100, 200))
        body_gradient.setColorAt(1, QColor(c.red() / 10 + 100, c.green() / 10 + 100, c.blue() / 10 + 100, 0))

        painter.setBrush(body_gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 12, 12)

        header_gradient = QLinearGradient(FlowTheme_Toy.get_header_rect(w, h, title_rect).topRight(),
                                          FlowTheme_Toy.get_header_rect(w, h, title_rect).bottomLeft())
        header_gradient.setColorAt(0, QColor(c.red(), c.green(), c.blue(), 255))
        header_gradient.setColorAt(1, QColor(c.red(), c.green(), c.blue(), 0))
        painter.setBrush(header_gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(FlowTheme_Toy.get_header_rect(w, h, title_rect), 12, 12)


    def draw_NI_small(self, node, painter: QPainter, c: QColor, w, h, bounding_rect, background_color=None):

        path = QPainterPath()
        path.moveTo(-w / 2, 0)

        path.cubicTo(-w / 2, -h / 2,
                     -w / 2, -h / 2,
                     0, -h / 2)
        path.cubicTo(+w / 2, -h / 2,
                     +w / 2, -h / 2,
                     +w / 2, 0)
        path.cubicTo(+w / 2, +h / 2,
                     +w / 2, +h / 2,
                     0, +h / 2)
        path.cubicTo(-w / 2, +h / 2,
                     -w / 2, +h / 2,
                     -w / 2, 0)
        path.closeSubpath()

        body_gradient = QLinearGradient(bounding_rect.bottomLeft(),
                                        bounding_rect.topRight())
        body_gradient.setColorAt(0, QColor(c.red(), c.green(), c.blue(), 150))
        body_gradient.setColorAt(1, QColor(c.red(), c.green(), c.blue(), 80))

        painter.setBrush(body_gradient)
        painter.setPen(QPen(QColor(30, 43, 48)))

        painter.drawPath(path)


class FlowTheme_DarkTron(FlowTheme):


    name = 'Tron'

    node_selection_stylesheet = ''

    header_padding = (12, 5, 10, 2)

    exec_conn_color = QColor(0, 120, 180)
    exec_conn_width = 4
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(0, 120, 180)
    data_conn_width = 4
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#333333'))

    
    def paint_NI_title_label(self, node, painter, option, hovering, node_style, node_title, node_color, node_item_bounding_rect):
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=node_color if not hovering else node_color.lighter(),
                pen_w=2,
                font=QFont('Poppins', 15),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=node_color,
                pen_w=2,
                font=QFont('K2D', 20, QFont.Bold, True),
                node_item_bounding_rect=node_item_bounding_rect
            )


    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        if type_ == 'exec':
            c = QColor('#FFFFFF')
        else:
            c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Source Code Pro", 10, QFont.Bold), bounding_rect)


    def paint_PI(self, node, painter, option, node_color, type_, connected, padding, w, h):

        color = QColor('#FFFFFF') if type_ == 'exec' else node_color
        pen = QPen(color)
        pen.setWidth(2)
        painter.setPen(pen)
        if connected or \
                option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
            r = node_color.red()
            g = node_color.green()
            b = node_color.blue()
            brush = QBrush(QColor(r, g, b, 100))
            painter.setBrush(brush)
        else:
            painter.setBrush(Qt.NoBrush)

        painter.drawEllipse(QRectF(padding, padding, w, h))


    def paint_NI(self, node, node_style,
                 painter, option,
                 color: QColor, w: int, h: int, bounding_rect, title_rect):

        painter.setRenderHint(QPainter.Antialiasing)

        if node_style == 'normal':
            self.draw_NI_normal(node, painter, color, w, h, bounding_rect, title_rect)
        elif node_style == 'small':
            if option.state & QStyle.State_MouseOver:  # use special dark background color when mouse hovers
                self.draw_NI_small(node, painter, color, w, h, bounding_rect, color.darker())
            else:
                self.draw_NI_small(node, painter, color, w, h, bounding_rect, QColor('#212429'))


    def draw_NI_normal(self, node, painter, c: QColor, w: int, h: int, bounding_rect, title_rect):

        background_color = QColor('#212224')
        painter.setBrush(background_color)
        pen = QPen(c)
        pen.setWidth(2)
        painter.setPen(pen)
        body_path = self.get_extended_body_path(w, h)
        painter.drawPath(body_path)

        header_gradient = QLinearGradient(self.get_header_rect(w, h, title_rect).topRight(),
                                          self.get_header_rect(w, h, title_rect).bottomLeft())
        header_gradient.setColorAt(0, QColor(c.red(), c.green(), c.blue(), 255))
        header_gradient.setColorAt(0.5, QColor(c.red(), c.green(), c.blue(), 100))
        header_gradient.setColorAt(1, QColor(c.red(), c.green(), c.blue(), 0))
        painter.setBrush(header_gradient)
        header_path = self.get_extended_header_path(w, h, title_rect)
        painter.drawPath(header_path)


    @staticmethod
    def get_extended_body_path(w, h):
        
        c_s = 10  # corner size

        path = QPainterPath()
        path.moveTo(+w / 2, -h / 2 + c_s)
        path.lineTo(+w / 2 - c_s, -h / 2)
        path.lineTo(-w / 2 + c_s, -h / 2)
        path.lineTo(-w / 2, -h / 2 + c_s)
        path.lineTo(-w / 2, +h / 2 - c_s)
        path.lineTo(-w / 2 + c_s, +h / 2)
        path.lineTo(+w / 2 - c_s, +h / 2)
        path.lineTo(+w / 2, +h / 2 - c_s)
        path.closeSubpath()
        return path

    
    def get_extended_header_path(self, w, h, title_rect):
        
        c_s = 10  # corner size

        # header_height = 35 * (NIPainter_DarkTron.ni.parent_node.title.count('\n') + 1)
        header_height = self.get_header_rect(w, h, title_rect).height()
        header_bottom = -h / 2 + header_height
        path = QPainterPath()
        path.moveTo(+w / 2, -h / 2 + c_s)
        path.lineTo(+w / 2 - c_s, -h / 2)
        path.lineTo(-w / 2 + c_s, -h / 2)
        path.lineTo(-w / 2, -h / 2 + c_s)
        path.lineTo(-w / 2, header_bottom - c_s)
        path.lineTo(-w / 2 + c_s, header_bottom)
        path.lineTo(+w / 2 - c_s, header_bottom)
        path.lineTo(+w / 2, header_bottom - c_s)
        path.closeSubpath()
        return path


    def draw_NI_small(self, node, painter: QPainter, c: QColor, w, h, bounding_rect, background_color=None):

        c_s = 10  # corner size

        path = QPainterPath()
        path.moveTo(-w / 2, 0)

        path.lineTo(-w / 2 + c_s / 2, -h / 2 + c_s / 2)
        path.lineTo(0, -h / 2)
        path.lineTo(+w / 2 - c_s / 2, -h / 2 + c_s / 2)
        path.lineTo(+w / 2, 0)
        path.lineTo(+w / 2 - c_s / 2, +h / 2 - c_s / 2)
        path.lineTo(0, +h / 2)
        path.lineTo(-w / 2 + c_s / 2, +h / 2 - c_s / 2)
        path.closeSubpath()

        painter.setBrush(background_color)
        pen = QPen(c)
        pen.setWidth(2)
        painter.setPen(pen)

        painter.drawPath(path)


class FlowTheme_Ghost(FlowTheme):


    name = 'Ghost'
    type_ = 'dark'

    node_selection_stylesheet = ''

    header_padding = (12, 6, 10, 2)

    exec_conn_color = QColor(0, 17, 25)
    exec_conn_width = 2
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(0, 17, 25)
    data_conn_width = 2
    data_conn_pen_style = Qt.DashLine

    flow_background_color = QColor('#333333')
    flow_background_brush = QBrush(flow_background_color)

    node_color = QColor(28, 28, 28, 170)
    node_small_color = QColor('#212429')

    EXPORT = [
        'nodes color',
        'small nodes color',
        'flow background color'
    ]


    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'nodes color':
                self.node_color = self.hex_to_col(v)
            elif k == 'small nodes color':
                self.node_small_color = self.hex_to_col(v)
            elif k == 'flow background color':
                self.flow_background_color = self.hex_to_col(v)
                self.flow_background_brush = QBrush(self.flow_background_color)


    def paint_NI_title_label(self, node, painter, option, hovering, node_style, node_title, node_color, 
                             node_item_bounding_rect):
        
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=node_color if not hovering else node_color.lighter(),
                pen_w=2,
                font=QFont('Poppins', 15),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=node_color,
                pen_w=2,
                font=QFont('K2D', 20, QFont.Bold, True),
                node_item_bounding_rect=node_item_bounding_rect
            )
    
    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        if type_ == 'exec':
            c = QColor('#FFFFFF')
        else:
            c = node_color
        
        self.paint_PI_label_default(painter, label_str, c, QFont("Source Code Pro", 10, QFont.Bold), bounding_rect)
    
    def paint_PI(self, node, painter, option, node_color, type_, connected, padding, w, h):

        color = QColor('#FFFFFF') if type_ == 'exec' else node_color

        if type_ == 'exec':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                brush = QBrush(QColor(255, 255, 255, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)
        elif type_ == 'data':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                r = node_color.red()
                g = node_color.green()
                b = node_color.blue()
                brush = QBrush(QColor(r, g, b, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

        pen = QPen(color)
        pen.setWidth(1)
        painter.setPen(pen)
        painter.drawEllipse(QRectF(padding+w/4, padding+h/4, w/2, h/2))


    def paint_NI(self, node, node_style,
                 painter, option,
                 color: QColor, w: int, h: int, bounding_rect, title_rect):

        painter.setRenderHint(QPainter.Antialiasing)

        if node_style == 'normal':
            self.draw_NI_normal(node, painter, color, w, h, bounding_rect, title_rect)
        elif node_style == 'small':
            if option.state & QStyle.State_MouseOver:  # use special dark background color when mouse hovers
                self.draw_NI_small(node, painter, color, w, h, bounding_rect, background_color=color.darker())
            else:
                self.draw_NI_small(node, painter, color, w, h, bounding_rect)


    def draw_NI_normal(self, node, painter, c, w, h, bounding_rect, title_rect):

        background_color = self.node_color

        # body_gradient = QRadialGradient(QPointF(bounding_rect.topLeft().x() + w,
        #                                         bounding_rect.topLeft().y() - h),
        #                                 pythagoras(2 * h, 2 * w))
        # body_gradient.setColorAt(0, QColor(c.red() / 10 + 100, c.green() / 10 + 100, c.blue() / 10 + 100, 100))
        # body_gradient.setColorAt(0.7, background_color)
        # body_gradient.setColorAt(1, background_color)
        #
        # painter.setBrush(body_gradient)

        painter.setBrush(background_color)
        pen = QPen(c.darker())
        pen.setWidth(1)
        painter.setPen(pen)
        body_path = self.get_extended_body_path(5, w, h)
        painter.drawPath(body_path)


    @staticmethod
    def get_extended_body_path(c_s, w, h):

        path = QPainterPath()
        path.moveTo(+w / 2, -h / 2 + c_s)
        path.lineTo(+w / 2 - c_s, -h / 2)
        path.lineTo(-w / 2 + c_s, -h / 2)
        path.lineTo(-w / 2, -h / 2 + c_s)
        path.lineTo(-w / 2, +h / 2 - c_s)
        path.lineTo(-w / 2 + c_s, +h / 2)
        path.lineTo(+w / 2 - c_s, +h / 2)
        path.lineTo(+w / 2, +h / 2 - c_s)
        path.closeSubpath()
        return path


    def draw_NI_small(self, node, painter, c, w, h, bounding_rect, background_color=None):

        background_color = self.node_small_color
        c_s = 10  # corner size
        
        path = self.get_extended_body_path(c_s, w, h)  # equals the small in this case

        painter.setBrush(background_color)
        # pen = QPen(QColor('#333333'))  # QPen(c)
        # pen.setWidth(1)
        # painter.setPen(pen)
        painter.setPen(Qt.NoPen)

        painter.drawPath(path)


class FlowTheme_Blender(FlowTheme):


    name = 'Blender'

    node_selection_stylesheet = ''

    header_padding = (8, 6, 10, 2)

    exec_conn_color = QColor(0, 17, 25)
    exec_conn_width = 2
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor(0, 17, 25)
    data_conn_width = 2
    data_conn_pen_style = Qt.DashLine

    flow_background_color = QColor('#333333')
    flow_background_brush = QBrush(flow_background_color)

    node_color = QColor(100, 100, 100, 150)

    EXPORT = [
        'nodes color',
        'flow background color'
    ]


    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'nodes color':
                self.node_color = self.hex_to_col(v)
            elif k == 'flow background color':
                self.flow_background_color = self.hex_to_col(v)
                self.flow_background_brush = QBrush(self.flow_background_color)


    def paint_NI_title_label(self, node, painter, option, hovering, node_style, node_title, node_color, node_item_bounding_rect):
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=QColor('#FFFFFF'),
                pen_w=2,
                font=QFont('Poppins', 15),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=node_color,
                pen_w=2,
                font=QFont('K2D', 20, QFont.Bold, True),
                node_item_bounding_rect=node_item_bounding_rect
            )


    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        if type_ == 'exec':
            c = QColor('#FFFFFF')
        else:
            c = node_color
        
        self.paint_PI_label_default(painter, label_str, c, QFont("Source Code Pro", 10, QFont.Bold), bounding_rect)


    def paint_PI(self, node, painter, option, node_color, type_, connected, padding, w, h):

        color = QColor('#FFFFFF') if type_ == 'exec' else node_color

        if type_ == 'exec':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                brush = QBrush(QColor(255, 255, 255, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)
        elif type_ == 'data':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                r = node_color.red()
                g = node_color.green()
                b = node_color.blue()
                brush = QBrush(QColor(r, g, b, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

        pen = QPen(color)
        pen.setWidth(1)
        painter.setPen(pen)
        painter.drawEllipse(QRectF(padding + w / 4, padding + h / 4, w / 2, h / 2))

    
    def draw_NI_normal(self, node, painter, c, w, h, bounding_rect, title_rect):

        background_color = self.node_color
        header_color = QColor(c.red(), c.green(), c.blue(), 180)

        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height+0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(QPen(c.darker()))
        painter.drawRoundedRect(bounding_rect, 7, 7)


    def draw_NI_small(self, node, painter, c, w, h, bounding_rect, background_color=None):

        background_color = QColor('#212429')
        c_s = 15
        painter.setBrush(self.interpolate_color(c, background_color, 0.97))
        painter.setPen(QPen(c))
        painter.drawRoundedRect(bounding_rect, c_s, c_s)


class FlowTheme_Simple(FlowTheme):


    name = 'Simple'
    type_ = 'dark'

    node_selection_stylesheet = ''

    header_padding = (10, 2, 10, 2)

    exec_conn_color = QColor('#989c9f')
    exec_conn_width = 2
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#989c9f')
    data_conn_width = 2
    data_conn_pen_style = Qt.DashLine

    flow_background_color = QColor('#3f4044')
    flow_background_brush = QBrush(flow_background_color)

    node_background_color = QColor('#212429')
    node_small_background_color = QColor('#212429')

    EXPORT = [
        'nodes background color',
        'small nodes background color',
        'flow background color'
    ]


    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'nodes background color':
                self.node_background_color = self.hex_to_col(v)
            elif k == 'small nodes background color':
                self.node_small_background_color = self.hex_to_col(v)
            elif k == 'flow background color':
                self.flow_background_color = self.hex_to_col(v)
                self.flow_background_brush = QBrush(self.flow_background_color)


    def paint_NI_title_label(self, node, painter, option, hovering, node_style, node_title, node_color, node_item_bounding_rect):
        if node_style == 'normal':
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=QColor('#312b29'),
                pen_w=2,
                font=QFont('ASAP', 13, QFont.Bold),
                node_item_bounding_rect=node_item_bounding_rect
            )
        else:
            self.paint_NI_title_label_default(
                painter=painter,
                node_style=node_style,
                title=node_title,
                color=node_color,
                pen_w=2,
                font=QFont('Poppins', 15, QFont.Thin),
                node_item_bounding_rect=node_item_bounding_rect
            )


    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = None
        if not connected:
            c = QColor('#53585c')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Courier New", 10, QFont.Bold), bounding_rect)


    def paint_PI(self, node, painter, option, node_color, type_, connected, padding, w, h):

        color = None
        if not connected:
            color = QColor('#53585c')
        else:
            if type_ == 'exec':
                color = QColor('#dddddd')
            else:
                color = node_color

        if type_ == 'exec':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                brush = QBrush(QColor(255, 255, 255, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)
        elif type_ == 'data':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                r = node_color.red()
                g = node_color.green()
                b = node_color.blue()
                brush = QBrush(QColor(r, g, b, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

        brush = QBrush(QColor(color))
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)

        painter.drawEllipse(QRectF(padding+w/8, padding+h/8, 3*w/4, 3*h/4))


    def draw_NI_normal(self, node, painter, c, w, h, bounding_rect, title_rect):

        background_color = self.node_background_color
        header_color = c

        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height+0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(Qt.NoPen)  # QPen(c.darker()))
        painter.drawRoundedRect(bounding_rect, 9, 9)


    def draw_NI_small(self, node, painter, c, w, h, bounding_rect, background_color=None):

        background_color = self.node_small_background_color
        c_s = 10
        painter.setBrush(self.interpolate_color(c, background_color, 0.97))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, c_s, c_s)


class FlowTheme_Ueli(FlowTheme):


    name = 'Ueli'

    node_selection_stylesheet = ''

    header_padding = (5, 2, 10, 0)

    exec_conn_color = QColor('#989c9f')
    exec_conn_width = 2
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#989c9f')
    data_conn_width = 2
    data_conn_pen_style = Qt.DashLine

    flow_background_color = QColor('#3f4044')
    flow_background_brush = QBrush(flow_background_color)

    nodes_background_color = QColor('#212429')
    small_nodes_background_color = nodes_background_color

    EXPORT = [
        'nodes background color',
        'small nodes background color',
        'flow background color'
    ]


    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'nodes background color':
                c = self.hex_to_col(v)
                self.nodes_background_color = c
            elif k == 'small nodes background color':
                self.small_nodes_background_color = self.hex_to_col(v)
            elif k == 'flow background color':
                self.flow_background_color = self.hex_to_col(v)
                self.flow_background_brush = QBrush(self.flow_background_color)


    def paint_NI_title_label(self, node, painter, option, hovering, node_style, node_title, node_color, node_item_bounding_rect):
        if node_style == 'normal':
            painter.setPen(QPen(QColor(node_color.name())))
            painter.setFont(QFont('Poppins', 13))
            painter.drawText(node_item_bounding_rect, Qt.AlignLeft | Qt.AlignVCenter, node_title)
        else:
            self.paint_NI_title_label_default(
                painter=painter, 
                node_style=node_style, 
                title=node_title, 
                color=node_color,
                pen_w=2,
                font=QFont('Poppins', 15, QFont.Thin),
                node_item_bounding_rect=node_item_bounding_rect
            )


    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        
        c = None
        if not connected:
            c = QColor('#53585c')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color
        
        self.paint_PI_label_default(painter, label_str, c, QFont("Courier New", 10, QFont.Bold), bounding_rect)


    def paint_PI(self, node, painter, option, node_color, type_, connected, padding, w, h):
        
        color = None
        if not connected:
            color = QColor('#53585c')
        else:
            if type_ == 'exec':
                color = QColor('#dddddd')
            else:
                color = node_color

        if type_ == 'exec':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                brush = QBrush(QColor(255, 255, 255, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)
        elif type_ == 'data':
            if connected or \
                    option.state & QStyle.State_MouseOver:  # also fill when mouse hovers
                r = node_color.red()
                g = node_color.green()
                b = node_color.blue()
                brush = QBrush(QColor(r, g, b, 100))
                painter.setBrush(brush)
            else:
                painter.setBrush(Qt.NoBrush)

        brush = QBrush(QColor(color))
        painter.setBrush(brush)
        painter.setPen(Qt.NoPen)

        painter.drawEllipse(QRectF(padding+w/8, padding+h/8, 3*w/4, 3*h/4))

    
    def draw_NI_normal(self, node, painter, c, w, h, bounding_rect: QRectF, title_rect):

        background_color = self.nodes_background_color
        header_color = c

        header_height = self.get_header_rect(w, h, title_rect).height()
        rel_header_height = header_height/h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height+0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(QBrush(background_color))
        painter.setPen(Qt.NoPen)  # QPen(c.darker()))
        painter.drawRoundedRect(QRectF(
            QPointF(bounding_rect.left(), bounding_rect.top()+header_height),
            bounding_rect.bottomRight()
        ), 6, 6)


    def draw_NI_small(self, node, painter, c, w, h, bounding_rect, background_color=None):
        background_color = self.small_nodes_background_color
        c_s = 10  # corner size
        painter.setBrush(self.interpolate_color(c, background_color, 0.97))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, c_s, c_s)


class FlowTheme_PureDark(FlowTheme):

    name = 'pure dark'

    node_selection_stylesheet = ''

    header_padding = (4, 2, 2, 2)

    exec_conn_color = QColor('#ffffff')
    exec_conn_width = 1.5
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#ffffff')
    data_conn_width = 1.5
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#1E242A'))
    flow_background_grid = ('points', flow_background_brush.color().lighter(), 2, 50, 50)

    node_item_shadow_color = QColor('#101010')

    node_ext_background_color = QColor('#0C1116')
    node_small_background_color = QColor('#363c41')
    node_title_color = QColor('#ffffff')
    port_pin_pen_color = QColor('#ffffff')

    EXPORT = [
        'extended node background color',
        'small node background color',
        'node title color',
        'port pin pen color'
    ]


    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'extended node background color':
                self.node_ext_background_color = self.hex_to_col(v)
            elif k == 'small node background color':
                self.node_small_background_color = self.hex_to_col(v)
            elif k == 'node title color':
                self.node_title_color = self.hex_to_col(v)
            elif k == 'port pin pen color':
                self.port_pin_pen_color = self.hex_to_col(v)


    def paint_NI_title_label(self, node, painter, option, hovering, node_style, node_title, node_color, node_item_bounding_rect):

        painter.setPen(QPen(self.node_title_color))

        if node_style == 'normal':
            painter.setFont(QFont('Segoe UI', 11))
            align = Qt.AlignLeft | Qt.AlignVCenter
        else:
            painter.setFont(QFont('Segoe UI', 15))
            align = Qt.AlignCenter

        painter.drawText(node_item_bounding_rect, align, node_title)


    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = None
        if not connected:
            c = QColor('#53585c')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color
        
        self.paint_PI_label_default(painter, label_str, c, QFont("Segoe UI", 10), bounding_rect)


    def paint_PI(self, node, painter, option, node_color, type_, connected, padding, w, h):

        if connected:
            painter.setBrush(QColor('#508AD8'))
            painter.setPen(Qt.NoPen)
        else:
            painter.setBrush(Qt.NoBrush)
            p = QPen(self.port_pin_pen_color)
            p.setWidthF(1.1)
            painter.setPen(p)

        rect = QRectF(padding + w / 8, padding + h / 8, 3 * w / 4, 3 * h / 4)
        if type_ == 'exec':
            painter.setBrush(QBrush(QColor('white')))
        painter.drawEllipse(rect)


    def draw_NI_normal(self, node, painter, c, w, h, bounding_rect: QRectF, title_rect):

        background_color = self.node_ext_background_color

        header_height = self.get_header_rect(w, h, title_rect).height()

        painter.setBrush(QBrush(background_color))
        painter.setPen(Qt.NoPen)  # QPen(c.darker()))
        painter.drawRoundedRect(QRectF(
            QPointF(bounding_rect.left(), bounding_rect.top() + header_height),
            bounding_rect.bottomRight()
        ), 3, 3)

        p = QPen(c)
        p.setWidthF(2.3)
        painter.setPen(p)
        painter.drawLine(
            QPointF(bounding_rect.left(), bounding_rect.top() + header_height),
            QPointF(bounding_rect.right(), bounding_rect.top() + header_height)
        )


    def draw_NI_small(self, node, painter, c, w, h, bounding_rect, background_color=None):
        
        painter.setBrush(QBrush(self.node_small_background_color))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 4, 4)


class FlowTheme_PureLight(FlowTheme_PureDark):

    name = 'pure light'
    type_ = 'light'

    header_padding = (2, 2, 2, 2)

    node_selection_stylesheet = '''
NodeSelectionWidget {
    background-color: white;
}
NodeWidget {
    background-color: white;
}
    '''

    exec_conn_color = QColor('#1f1f1f')

    data_conn_color = QColor('#1f1f1f')

    flow_background_brush = QBrush(QColor('#ffffff'))
    flow_background_grid = ('points', QColor('#dddddd'), 2, 20, 20)


    node_ext_background_color = QColor('#cdcfd1')
    node_small_background_color = QColor('#bebfc1')
    node_title_color = QColor('#1f1f1f')
    port_pin_pen_color = QColor('#1f1f1f')

    node_item_shadow_color = QColor('#cccccc')


class FlowTheme_Colorful(FlowTheme):

    name = 'colorful dark'

    header_padding = (12, 0, 2, 2)

    exec_conn_color = QColor('#ffffff')
    exec_conn_width = 1.5
    exec_conn_pen_style = Qt.SolidLine

    data_conn_color = QColor('#ffffff')
    data_conn_width = 1.5
    data_conn_pen_style = Qt.DashLine

    flow_background_brush = QBrush(QColor('#1E242A'))
    flow_background_grid = ('points', flow_background_brush.color().lighter(), 2, 50, 50)


    node_ext_background_color = QColor('#0C1116')
    node_small_background_color = QColor('#363c41')
    node_title_color = QColor('#ffffff')
    port_pin_pen_color = QColor('#ffffff')

    EXPORT = [
        'node title color',
        'port pin pen color'
    ]


    def _load(self, imported: dict):
        super()._load(imported)

        for k, v in imported.items():
            if k == 'node title color':
                self.node_title_color = self.hex_to_col(v)
            elif k == 'port pin pen color':
                self.port_pin_pen_color = self.hex_to_col(v)


    def paint_NI_title_label(self, node, painter, option, hovering, node_style, node_title, node_color,
                             node_item_bounding_rect):

        painter.setPen(QPen(self.node_title_color))

        if node_style == 'normal':
            painter.setFont(QFont('Segoe UI', 11))
            align = Qt.AlignLeft | Qt.AlignVCenter
        else:
            painter.setFont(QFont('Segoe UI', 15))
            align = Qt.AlignCenter

        painter.drawText(node_item_bounding_rect, align, node_title)


    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = None
        if not connected:
            c = QColor('#dddddd')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Segoe UI", 10), bounding_rect)


    def paint_PI(self, node, painter, option, node_color, type_, connected, padding, w, h):

        if connected:
            painter.setBrush(QColor('#508AD8'))
            painter.setPen(Qt.NoPen)
        else:
            painter.setBrush(Qt.NoBrush)
            p = QPen(self.port_pin_pen_color)
            p.setWidthF(1.1)
            painter.setPen(p)

        painter.drawEllipse(QRectF(padding + w / 8, padding + h / 8, 3 * w / 4, 3 * h / 4))


    def draw_NI_normal(self, node, painter, c, w, h, bounding_rect: QRectF, title_rect):

        background_color = c
        background_color.setAlpha(150)

        header_color = QColor(c.red(), c.green(), c.blue(), 130).darker()

        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height+0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 7, 7)

    def draw_NI_small(self, node, painter, c, w, h, bounding_rect, background_color=None):

        painter.setBrush(QBrush(QColor(c.red(), c.green(), c.blue(), 150)))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 8, 8)


class FlowTheme_ColorfulLight(FlowTheme_Colorful):

    name = 'colorful light'
    type_ = 'light'

    header_padding = (12, 0, 2, 2)

    exec_conn_color = QColor('#1f1f1f')

    data_conn_color = QColor('#1f1f1f')

    flow_background_brush = QBrush(QColor('#ffffff'))
    flow_background_grid = ('points', QColor('#dddddd'), 2, 20, 20)

    node_title_color = QColor('#1f1f1f')
    port_pin_pen_color = QColor('#1f1f1f')

    node_item_shadow_color = QColor('#cccccc')


    def paint_PI_label(self, node, painter, option, type_, connected, label_str, node_color, bounding_rect):
        c = None
        if not connected:
            c = QColor('#1f1f1f')
        else:
            if type_ == 'exec':
                c = QColor('#cccccc')
            else:
                c = node_color

        self.paint_PI_label_default(painter, label_str, c, QFont("Segoe UI", 10), bounding_rect)


    def draw_NI_normal(self, node, painter, c, w, h, bounding_rect: QRectF, title_rect):

        background_color = c.lighter()
        background_color.setAlpha(150)

        header_color = QColor(c.red(), c.green(), c.blue(), 130).darker()

        rel_header_height = self.get_header_rect(w, h, title_rect).height() / h
        gradient = QLinearGradient(bounding_rect.topLeft(), bounding_rect.bottomLeft())
        gradient.setColorAt(0, header_color)
        gradient.setColorAt(rel_header_height, header_color)
        gradient.setColorAt(rel_header_height+0.0001, background_color)
        gradient.setColorAt(1, background_color)

        painter.setBrush(gradient)
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 7, 7)


    def draw_NI_small(self, node, painter, c, w, h, bounding_rect, background_color=None):

        painter.setBrush(QBrush(QColor(c.red(), c.green(), c.blue(), 150)))
        painter.setPen(Qt.NoPen)
        painter.drawRoundedRect(bounding_rect, 8, 8)



from qtpy.QtCore import QSize, QRectF, QPointF, QSizeF
from qtpy.QtGui import QPixmap, QImage, QPainter, QIcon, QPicture
from qtpy.QtWidgets import QGraphicsPixmapItem, QGraphicsWidget, QGraphicsLayoutItem

from .tools import change_svg_color


class NodeItem_Icon(QGraphicsWidget):
    def __init__(self, node, node_item):
        super().__init__(parent=node_item)

        if node.style == 'normal':
            self.size = QSize(20, 20)
        else:
            self.size = QSize(50, 50)

        self.setGraphicsItem(self)

        image = QImage(node.icon)
        self.pixmap = QPixmap.fromImage(image)
        # self.pixmap = change_svg_color(node.icon, node.color)


    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.size)

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())

    def sizeHint(self, which, constraint=...):
        return QSizeF(self.size.width(), self.size.height())


    def paint(self, painter, option, widget=None):

        # TODO: anti aliasing for node icons

        # this doesn't work: ...
        # painter.setRenderHint(QPainter.Antialiasing, True)
        # painter.setRenderHint(QPainter.HighQualityAntialiasing, True)
        # painter.setRenderHint(QPainter.SmoothPixmapTransform, True)


        painter.drawPixmap(
            0, 0,
            self.size.width(), self.size.height(),
            self.pixmap
        )



from qtpy.QtWidgets import QGraphicsProxyWidget


class FlowViewProxyWidget(QGraphicsProxyWidget):
    """Ensures easy controls event handling for QProxyWidgets in the flow."""

    def __init__(self, flow_view, parent=None):
        super(FlowViewProxyWidget, self).__init__(parent)

        self.flow_view = flow_view


    def mousePressEvent(self, arg__1):
        QGraphicsProxyWidget.mousePressEvent(self, arg__1)
        if arg__1.isAccepted():
            self.flow_view.mouse_event_taken = True

    def mouseReleaseEvent(self, arg__1):
        self.flow_view.mouse_event_taken = True
        QGraphicsProxyWidget.mouseReleaseEvent(self, arg__1)

    def wheelEvent(self, event):
        QGraphicsProxyWidget.wheelEvent(self, event)

    def keyPressEvent(self, arg__1):
        QGraphicsProxyWidget.keyPressEvent(self, arg__1)
        if arg__1.isAccepted():
            self.flow_view.ignore_key_event = True


import enum
import json
import os
from math import sqrt

from .ryvencore.tools import serialize, deserialize


def pythagoras(a, b):
    return sqrt(a ** 2 + b ** 2)


def get_longest_line(s: str):
    lines = s.split('\n')
    lines = [line.replace('\n', '') for line in lines]
    longest_line_found = ''
    for line in lines:
        if len(line) > len(longest_line_found):
            longest_line_found = line
    return line


def shorten(s: str, max_chars: int, line_break: bool = False):
    """Ensures, that a given string does not exceed a given max length. If it would, its cut in the middle."""
    l = len(s)
    if l > max_chars:
        insert = ' . . . '
        if line_break:
            insert = '\n'+insert+'\n'
        insert_length = len(insert)
        left = s[:round((max_chars-insert_length)/2)]
        right = s[round(l-((max_chars-insert_length)/2)):]
        return left+insert+right
    else:
        return s


class MovementEnum(enum.Enum):
    # this should maybe get removed later
    mouse_clicked = 1
    position_changed = 2
    mouse_released = 3


def translate_project(filepath: str):
    """Performs a few minor changes to a project file using the structure from Ryven 2, to make it compatible
    with the current system"""

    f = open(filepath, 'r')
    project_str = f.read()
    f.close()

    project_str = project_str.replace('flow widget config', 'flow view')
    project_str = project_str.replace('widget position', 'widget pos')

    project: dict = json.loads(project_str)
    project_repaired = translate_project__repair(project)
    project_repaired_str = json.dumps(project_repaired, indent=4)

    f = open(os.path.splitext(filepath)[0]+'_TRANSLATED.rpo', 'w')
    f.write(project_repaired_str)
    f.close()


def translate_project__repair(obj):
    if type(obj) == dict:
        for k, v in obj.items():
            if k in ('state data', 'widget data', 'main widget data'):
                obj[k] = serialize(v)

            # elif k == 'flow widget config':
            #     obj['flow view'] = obj['flow widget config']
            #     del obj['flow widget config']

            elif k == 'widget pos' and v == 'under':
                obj[k] = 'below'

            else:
                obj[k] = translate_project__repair(v)

    elif type(obj) == list:
        for i in range(len(obj)):
            obj[i] = translate_project__repair(obj[i])

    return obj


def change_svg_color(filepath: str, color_hex: str):


    from qtpy.QtSvg import QSvgRenderer
    from qtpy.QtGui import QPixmap, QPainter, Qt

    with open(filepath) as f:
        data = f.read()
    data = data.replace('fill:#000000', 'fill:'+color_hex)
    with open(filepath, 'w') as f:
        f.write(data)

    svg_renderer = QSvgRenderer(filepath)
    pix = QPixmap(svg_renderer.defaultSize())
    pix.fill(Qt.transparent)
    pix_painter = QPainter(pix)
    svg_renderer.render(pix_painter)

    return pix



    # """
    # Changes the color of an SVG image and returns a QPixmap
    #
    # https://stackoverflow.com/questions/15123544/change-the-color-of-an-svg-in-qt
    # """
    #
    # from qtpy.QtGui import Qt, QPainter
    # from qtpy.QtXml import QDomDocument, QDomElement
    # from qtpy.QtSvg import QSvgRenderer
    # from qtpy.QtGui import QPixmap
    #
    #
    # def change_svg_color__set_attr_recur(elem: QDomElement, strtagname: str, strattr: str, strattrval: str):
    #
    #     # if it has the tag name then overwrite desired attribute
    #     if elem.tagName() == strtagname:
    #         elem.setAttribute(strattr, strattrval)
    #
    #     # loop all children
    #     for i in range(elem.childNodes().count()):
    #         if not elem.childNodes().at(i).isElement():
    #             continue
    #
    #         change_svg_color__set_attr_recur(elem.childNodes().at(i).toElement(), strtagname, strattr, strattrval)
    #
    #
    # # open svg resource load contents to qbytearray
    # f = open(filepath)
    # data = f.read()
    # f.close()
    # # load svg contents to xml document and edit contents
    # doc = QDomDocument()
    # doc.setContent(data)
    # # recursively change color
    # change_svg_color__set_attr_recur(doc.documentElement(), 'path', 'fill', color_hex)
    # # create svg renderer with edited contents
    # svg_renderer = QSvgRenderer(doc.toByteArray())
    # # create pixmap target (could be a QImage)
    # pix = QPixmap(svg_renderer.defaultSize())
    # pix.fill(Qt.transparent)
    # # create painter to act over pixmap
    # pix_painter = QPainter(pix)
    # # use renderer to render over painter which paints on pixmap
    # svg_renderer.render(pix_painter)
    #
    # return pix


from .Base import Base
from .Script import Script
from .RC import CLASSES
from .MacroNodeTypes import build_macro_classes


class MacroScript(Script):
    """Besides all the properties of a Script, a MacroScript automatically create an input node and an output node
    in the Flow, and locally defines a new Node class and registers it in the session."""

    id_ctr = Base.IDCtr()  # using custom ID to count MacroScripts
    # this ensures that I don't need to change the macro node class's identifier when the script's title changes.
    # the ID gets stored when serializing and is reloaded in a way that the ctr will afterwards still be larger than
    # any ID used, even if there were macro scripts deleted which created holes in the id assignments, so we don't get
    # dangerous overlapping

    MacroInputNode, MacroOutputNode, MacroNode = None, None, None

    @classmethod
    def build_node_classes(cls):
        cls.MacroInputNode, \
        cls.MacroOutputNode, \
        cls.MacroNode = \
            build_macro_classes(BaseClass=CLASSES['node base'])


    def __init__(self, session, title: str = None, config_data: dict = None, create_default_logs=True):

        self.input_node, self.output_node = None, None
        self.parameters: [dict] = []
        self.returns: [dict] = []
        # self.caller_stack = []  # : [FunctionScriptNode]; used by input, output and function nodes
        self.caller = None

        self.ID = self.id_ctr.count()
        if config_data and 'ID' in config_data:
            self.ID = max(config_data['ID'], self.ID)
        self.id_ctr.set_count(self.ID)

        super().__init__(session, title, config_data, create_default_logs)

        _self = self

        class CustomMacrotionScriptNode(_self.MacroNode):

            identifier = _self.macro_node_identifier()

            title = _self.title

            # it is important that these two static attributes are defined here ant NOT in MacroScriptNode
            # otherwise they would result in the same for ALL MacroScriptNodes
            macro_script = _self
            instances = []

        self.macro_node_class = CustomMacrotionScriptNode
        self.session.register_node(self.macro_node_class)

        if self.init_config:
            self.parameters = self.init_config['parameters']
            self.returns = self.init_config['returns']


    def macro_node_identifier(self):
        # not using the script title here anymore, because although it's unique, it can change
        return f'MACRO_NODE_{self.ID}'


    def load_flow(self):
        super().load_flow()

        if self.init_config:
            # find input and output node that have already been created by the flow
            for node in self.flow.nodes:
                if node.identifier == self.MacroInputNode.identifier:
                    self.input_node = node
                elif node.identifier == self.MacroOutputNode.identifier:
                    self.output_node = node
        else:
            self.input_node = self.flow.create_node(self.MacroInputNode)
            self.output_node = self.flow.create_node(self.MacroOutputNode)


    def add_parameter(self, type_, label):
        self.parameters.append({'type': type_, 'label': label})

        for mn in self.macro_node_class.instances:
            mn.create_input(label=label, type_=type_)

    def remove_parameter(self, index):
        self.parameters.remove(self.parameters[index])

        for mn in self.macro_node_class.instances:
            mn.delete_input(index)

    def add_return(self, type_, label):
        self.returns.append({'type': type_, 'label': label})

        for mn in self.macro_node_class.instances:
            mn.create_output(label, type_)

    def remove_return(self, index):
        self.returns.remove(self.returns[index])

        for mn in self.macro_node_class.instances:
            mn.delete_output(index)

    def serialize(self) -> dict:
        script_dict = super().serialize()

        script_dict['parameters'] = self.parameters
        script_dict['returns'] = self.returns
        script_dict['ID'] = self.ID

        return script_dict



"""
This module contains definitions of declarative types which can be conveniently used for data inputs.
For each of the exposed types, the frontend can define some prebuilt widgets, which makes nodes
from other frontends that use those types to determine input widgets automatically compatible in
other environments.
This list may grow significantly over time.
"""


class DType:
    def __init__(self, default, bounds: tuple = None, doc: str = "", _load_state=None):
        if _load_state:
            self.set_state(_load_state)
        else:
            self.default = default
            self.doc = doc
            self.bounds = bounds

        self._data = ['default', 'doc', 'bounds']

    def __str__(self):
        return 'DType.'+self.__class__.__name__

    @staticmethod
    def from_str(s):
        for DTypeClass in dtypes:
            if s == 'DType.'+DTypeClass.__name__:
                return DTypeClass

        return None

    def add_data(self, *attr_names):
        self._data += list(attr_names)

    def get_state(self) -> dict:
        return {
            name: getattr(self, name)
            for name in self._data
        }

    def set_state(self, data: dict):
        for name, val in data.items():
            setattr(self, name, val)


class Data(DType):
    """Any kind of data represented by some evaluated text input"""
    def __init__(self, default=None, size: str = 'm', doc: str = "", _load_state=None):
        """
        size: 's' / 'm' / 'l'
        """
        self.size = size
        super().__init__(default=default, doc=doc, _load_state=_load_state)
        self.add_data('size')


class Integer(DType):
    def __init__(self, default: int = 0, bounds: tuple = None, doc: str = "", _load_state=None):
        super().__init__(default=default, bounds=bounds, doc=doc, _load_state=_load_state)


class Float(DType):
    def __init__(self, default: float = 0.0, bounds: tuple = None, decimals: int = 10, doc: str = "", _load_state=None):
        self.decimals = decimals
        super().__init__(default=default, bounds=bounds, doc=doc, _load_state=_load_state)
        self.add_data('decimals')


class Boolean(DType):
    def __init__(self, default: bool = False, doc: str = "", _load_state=None):
        super().__init__(default=default, doc=doc, _load_state=_load_state)


class Char(DType):
    def __init__(self, default: chr = '', doc: str = "", _load_state=None):
        super().__init__(default=default, doc=doc, _load_state=_load_state)


class String(DType):
    def __init__(self, default: str = "", size: str = 'm', doc: str = "", _load_state=None):
        """
        size: 's' / 'm' / 'l'
        """
        self.size = size
        super().__init__(default=default, doc=doc, _load_state=_load_state)
        self.add_data('size')


class Choice(DType):
    def __init__(self, default, items: list, doc: str = "", _load_state=None):
        self.items = items
        super().__init__(default=default, doc=doc, _load_state=_load_state)
        self.add_data('items')


class List(DType):
    def __init__(self, default: list = [], doc: str = "", _load_state=None):
        super().__init__(default=default, doc=doc, _load_state=_load_state)


class Date(DType):
    ...


class Time(DType):
    ...


class Color(DType):
    ...


class Range(DType):
    ...

dtypes = [Data, Integer, Float, Boolean, Char, String, Choice, List]



from ..Base import Base

from .Variable import Variable

from ..InfoMsgs import InfoMsgs


class VarsManager(Base):
    """Manages script variables and triggers receivers when values of variables change"""

    def __init__(self, script, config=None):
        Base.__init__(self)

        self.script = script

        self.variables: [Variable] = []
        self.var_receivers = {}

        if config is not None:
            for name in config.keys():  # variables
                self.create_new_var(name, val=config[name])

    def var_name_valid(self, name: str) -> bool:
        """Checks if a var name candidate is empty or already used"""

        if len(name) == 0:
            return False

        # search for name issues
        for v in self.variables:
            if v.name == name:
                return False

        return True

    def create_new_var(self, name: str, val=None) -> Variable:
        """Creates and returns a new script variable"""

        if self.var_name_valid(name):
            v = Variable(name, val)
            self.variables.append(v)
            self.var_receivers[name] = {}
            return v
        else:
            return None

    def get_var(self, name) -> Variable:
        """Returns script variable with given name or None if it couldn't be found."""

        InfoMsgs.write('getting variable with name:', name)

        for v in self.variables:
            if v.name == name:
                return v
        return None

    def get_var_val(self, name):
        """Returns the value of a script variable with given name or None if it couldn't be found"""

        var = self.get_var(name)
        return var.val if var is not None else None

    def set_var(self, name, val) -> bool:
        """Sets the value of an existing script variable.
        Returns true in case of success, false if the var couldn't be found and set."""

        var_index = self._get_var_index_from_name(name)
        if var_index is None:
            return False

        var = self.variables[var_index]
        var.val = val

        # update all variable usages by calling all registered object's methods on updated variable with the new val
        for receiver, methods in self.var_receivers[name].items():
            for m in methods:
                m(name, val)

        # for receiver, var_name in self.var_receivers.keys():
        #     if var_name == name:
        #         self.var_receivers[receiver, var_name](var_name, val)  # calling the slot method

        return True

    def _get_var_index_from_name(self, name):
        var_names_list = [v.name for v in self.variables]
        for i in range(len(var_names_list)):
            if var_names_list[i] == name:
                return i

        return None

    def delete_var(self, var: Variable):
        """Deletes a variable."""

        self.variables.remove(var)

    def register_receiver(self, receiver, var_name: str, method):
        """A registered receiver (method) gets triggered every time the
        value of a variable with the given name changes (also when it gets created)."""

        # self.var_receivers[(receiver, var_name)] = method

        if receiver in self.var_receivers[var_name]:
            self.var_receivers[var_name][receiver].append(method)
        else:
            self.var_receivers[var_name][receiver] = [method]



    def unregister_receiver(self, receiver, var_name: str, method) -> bool:
        """Unregisters a method and returns true in case of success. See also register_receiver()."""

        if receiver in self.var_receivers[var_name]:
            self.var_receivers[var_name][receiver].remove(method)
            return True
        else:
            return False

        # try:
        #     del self.var_receivers[(receiver, var_name)]
        #     return True
        # except Exception:
        #     return False

    def config_data(self) -> dict:
        """Returns the config data of the script variables."""

        vars_dict = {}
        for v in self.variables:
            vars_dict[v.name] = {'serialized': v.serialize()}
        return vars_dict



from setuptools import setup


setup(
    name='ryvencore-qt',
    version='0.0.1',
    license='LGPLv2.1',
    description='Library for building Visual Scripting Editors with Qt',
    author='Leon Thomm',
    author_email='l.thomm@mailbox.org',
    packages=[
        'ryvencore_qt',
        'ryvencore_qt.resources',

        'ryvencore_qt.src',
        'ryvencore_qt.src.conv_gui',
        'ryvencore_qt.src.node_selection_widget',

        'ryvencore_qt.src.ryvencore',
        'ryvencore_qt.src.ryvencore.logging',
        'ryvencore_qt.src.ryvencore.script_variables',
        'ryvencore_qt.src.ryvencore.dtypes',
    ],
    include_package_data=True,
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "License :: OSI Approved :: GNU Lesser General Public License v2 or later (LGPLv2+)",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
    ],
    python_requires='>=3.9',
    install_requires=['PySide2', 'QtPy', 'Jinja2', 'Pygments'],
)



from qtpy.QtWidgets import QDialog, QDialogButtonBox, QVBoxLayout, QPlainTextEdit, QShortcut, QMessageBox
from qtpy.QtGui import QKeySequence


class EditVal_Dialog(QDialog):

    def __init__(self, parent, init_val):
        super(EditVal_Dialog, self).__init__(parent)

        # shortcut
        save_shortcut = QShortcut(QKeySequence.Save, self)
        save_shortcut.activated.connect(self.save_triggered)

        main_layout = QVBoxLayout()

        self.val_text_edit = QPlainTextEdit()
        val_str = ''
        try:
            val_str = str(init_val)
        except Exception as e:
            msg_box = QMessageBox(QMessageBox.Warning, 'Value parsing failed',
                                  'Couldn\'t stringify value', QMessageBox.Ok, self)
            msg_box.setDefaultButton(QMessageBox.Ok)
            msg_box.exec_()
            self.reject()

        self.val_text_edit.setPlainText(val_str)

        main_layout.addWidget(self.val_text_edit)

        button_box = QDialogButtonBox()
        button_box.setStandardButtons(QDialogButtonBox.Cancel | QDialogButtonBox.Ok)
        button_box.accepted.connect(self.accept)
        button_box.rejected.connect(self.reject)

        main_layout.addWidget(button_box)

        self.setLayout(main_layout)
        self.resize(450, 300)

        self.setWindowTitle('edit val')

    def save_triggered(self):
        self.accept()


    def get_val(self):
        val = self.val_text_edit.toPlainText()
        try:
            val = eval(val)
        except Exception as e:
            pass
        return val


class Base:

    class IDCtr:
        def __init__(self):
            self.ctr = -1

        def count(self):
            """increases the counter and returns the new count. first time is 0"""
            self.ctr += 1
            return self.ctr

        def set_count(self, cnt):
            if cnt < self.ctr:
                raise Exception("Decreasing ID counters is illegal")
            else:
                self.ctr = cnt

    # all abstract components have a global ID
    global_id_ctr = IDCtr()

    # optional custom ID counter
    id_ctr = None

    def __init__(self):
        self.GLOBAL_ID = self.global_id_ctr.count()

        # ignore custom ID if it has already been set
        if self.id_ctr is not None and not (hasattr(self, 'ID') and self.ID is not None):
            self.ID = self.id_ctr.count()



from qtpy.QtCore import QRectF, QPointF, QSizeF, Property
from qtpy.QtGui import QFont, QFontMetricsF, QColor
from qtpy.QtWidgets import QGraphicsWidget, QGraphicsLayoutItem, QGraphicsItem

from .tools import get_longest_line


class TitleLabel(QGraphicsWidget):

    # TODO: add NodeItem renaming on double click

    def __init__(self, node, node_item):
        super(TitleLabel, self).__init__(parent=node_item)

        self.setGraphicsItem(self)

        self.node = node
        self.node_item = node_item

        font = QFont('Poppins', 15) if self.node.style == 'normal' else \
            QFont('K2D', 20, QFont.Bold, True)  # should be quite similar to every specific font chosen by the painter
        self.fm = QFontMetricsF(font)
        self.title_str, self.width, self.height = None, None, None
        self.update_shape()

        self.color = QColor(30, 43, 48)
        self.pen_width = 1.5
        self.hovering = False  # whether the mouse is hovering over the parent NI (!)

        # # Design.flow_theme_changed.connect(self.theme_changed)
        # self.update_design()

    def update_shape(self):
        self.title_str = self.node.display_title

        # approximately!
        self.width = self.fm.width(get_longest_line(self.title_str)+'___')
        self.height = self.fm.height() * 0.7 * (self.title_str.count('\n') + 1)

    def boundingRect(self):
        return QRectF(QPointF(0, 0), self.geometry().size())

    def setGeometry(self, rect):
        self.prepareGeometryChange()
        QGraphicsLayoutItem.setGeometry(self, rect)
        self.setPos(rect.topLeft())

    def sizeHint(self, which, constraint=...):
        return QSizeF(self.width, self.height)

    def paint(self, painter, option, widget=None):

        self.node_item.session_design.flow_theme.paint_NI_title_label(
            self.node,
            painter, option, self.hovering,
            self.design_style(), self.title_str, self.node_item.color,
            self.boundingRect()
        )

    def design_style(self):
        return self.node.style

    def set_NI_hover_state(self, hovering: bool):
        self.hovering = hovering
        # self.update_design()
        self.update()

    # ANIMATION STUFF
    def get_color(self):
        return self.color

    def set_color(self, val):
        self.color = val
        QGraphicsItem.update(self)

    p_color = Property(QColor, get_color, set_color)


from .dtypes import DType


class NodePortBP:
    """
    The NodePorts (NodeInputs and NodeOutputs) are only placeholders for the static init_input and init_outputs of
    custom Node classes.
    An instantiated Node's actual inputs and outputs will be of type NodeObjPort (NodeObjInput, NodeObjOutput).
    """

    def __init__(self, label: str = '', type_: str = 'data'):
        """
        :type_: 'data' or 'exec'
        """

        self.type_: str = type_
        self.label: str = label


class NodeInputBP(NodePortBP):
    def __init__(self, label: str = '', type_: str = 'data', dtype: DType = None, add_config={}):

        super().__init__(label, type_)

        self.dtype = dtype
        self.add_config = add_config


class NodeOutputBP(NodePortBP):
    def __init__(self, label: str = '', type_: str = 'data'):
        super().__init__(label, type_)



from qtpy.QtWidgets import QWidget, QPushButton, QHBoxLayout, QSlider, QColorDialog
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor


class FlowViewStylusModesWidget(QWidget):
    def __init__(self, flow_view):
        super(FlowViewStylusModesWidget, self).__init__()

        self.setObjectName('FlowViewStylusModesWidget')

        # GENERAL ATTRIBUTES
        self.flow_view = flow_view
        self.pen_color = QColor(255, 255, 0)
        self.stylus_buttons_visible = True

        # stylus button
        self.stylus_button = QPushButton('stylus')  # show/hide
        self.stylus_button.clicked.connect(self.on_stylus_button_clicked)

        # mode
        self.set_stylus_mode_comment_button = QPushButton('comment')
        self.set_stylus_mode_comment_button.clicked.connect(self.on_comment_button_clicked)
        self.set_stylus_mode_edit_button = QPushButton('edit')
        self.set_stylus_mode_edit_button.clicked.connect(self.on_edit_button_clicked)

        # pen style
        self.pen_color_button = QPushButton('color')
        self.pen_color_button.clicked.connect(self.on_choose_color_clicked)
        self.pen_width_slider = QSlider(Qt.Horizontal)
        self.pen_width_slider.setRange(1, 100)
        self.pen_width_slider.setValue(20)


        # MAIN LAYOUT
        main_horizontal_layout = QHBoxLayout()

        main_horizontal_layout.addWidget(self.pen_color_button)
        main_horizontal_layout.addWidget(self.pen_width_slider)
        main_horizontal_layout.addWidget(self.set_stylus_mode_comment_button)
        main_horizontal_layout.addWidget(self.set_stylus_mode_edit_button)
        main_horizontal_layout.addWidget(self.stylus_button)

        self.setLayout(main_horizontal_layout)

        self.setStyleSheet('''
        QWidget#FlowViewStylusModesWidget {
            background: transparent; 
        }
                ''')

        self.hide_stylus_buttons()
        self.hide_pen_style_widgets()


    def hide_stylus_buttons(self):
        self.set_stylus_mode_edit_button.hide()
        self.set_stylus_mode_comment_button.hide()
        self.stylus_buttons_visible = False

    def show_stylus_buttons(self):
        self.set_stylus_mode_edit_button.show()
        self.set_stylus_mode_comment_button.show()
        self.stylus_buttons_visible = True

    def hide_pen_style_widgets(self):
        self.pen_color_button.hide()
        self.pen_width_slider.hide()

    def show_pen_style_widgets(self):
        self.pen_color_button.show()
        self.pen_width_slider.show()

    def on_stylus_button_clicked(self):
        if self.stylus_buttons_visible:
            self.hide_pen_style_widgets()
            self.hide_stylus_buttons()
        else:
            self.show_stylus_buttons()

        self.adjustSize()
        self.flow_view.set_stylus_proxy_pos()

    def on_edit_button_clicked(self):
        self.flow_view.stylus_mode = 'edit'
        # self.pen_style_widget.hide()
        self.hide_pen_style_widgets()

        # if I don't hide and show the settings_widget manually here, the stylus mode buttons take up the additional
        # space when clicking on comment and then edit. self.adjustSize() does not seem to work properly here...
        self.hide_stylus_buttons()
        self.show_stylus_buttons()
        # self.settings_widget.hide()
        # self.settings_widget.show()

        self.adjustSize()
        self.flow_view.set_stylus_proxy_pos()
        # self.flow.setDragMode(QGraphicsView.RubberBandDrag)

    def on_comment_button_clicked(self):
        self.flow_view.stylus_mode = 'comment'
        # self.pen_style_widget.show()
        self.show_pen_style_widgets()
        self.adjustSize()
        self.flow_view.set_stylus_proxy_pos()
        # self.flow.setDragMode(QGraphicsView.NoDrag)

    def on_choose_color_clicked(self):
        self.pen_color = QColorDialog.getColor(self.pen_color, options=QColorDialog.ShowAlphaChannel,
                                               title='Choose pen color')
        self.update_color_button_SS()


    def update_color_button_SS(self):

        self.pen_color_button.setStyleSheet(
            '''
QPushButton {
    background-color: '''+self.pen_color.name()+''';
}'''
        )

    def get_pen_settings(self):
        return {'color': self.pen_color.name(),
                'base stroke weight': self.pen_width_slider.value()/10}


import time

from qtpy.QtWidgets import QWidget, QApplication
from qtpy.QtCore import QObject, Signal, Qt

from .ryvencore import Session as RC_Session, Script
from .ryvencore.MacroScript import MacroScript
from .ryvencore.RC import CLASSES

from .Design import Design
from .ConnectionItem import DataConnectionItem, ExecConnectionItem
from .SessionThreadingBridge import SessionThreadingBridge
from .Node import Node
from .FlowView import FlowView
from .WRAPPERS import VarsManager, LogsManager, DataConnection, Flow, Logger


class Session(RC_Session, QObject):

    new_script_created = Signal(object)
    flow_view_created = Signal(object, object)
    script_renamed = Signal(object)
    script_deleted = Signal(object)

    build_flow_view_request = Signal(object, tuple)
    complete_flow_view_config = Signal(dict)

    def __init__(
            self,
            threaded: bool = False,
            gui_parent: QWidget = None,
            flow_theme_name=None,
            performance_mode=None,
            data_conn_class=None,
            data_conn_item_class=None,
            exec_conn_class=None,
            exec_conn_item_class=None,
            node_class=None,
    ):
        QObject.__init__(self)

        # custom WRAPPERS
        CLASSES['node base'] = Node if not node_class else node_class
        CLASSES['data conn'] = DataConnection if not data_conn_class else data_conn_class
        CLASSES['logs manager'] = LogsManager
        CLASSES['logger'] = Logger
        CLASSES['vars manager'] = VarsManager
        CLASSES['flow'] = Flow

        if exec_conn_class:
            CLASSES['exec conn'] = exec_conn_class

        # register additional classes
        CLASSES['data conn item'] = DataConnectionItem if not data_conn_item_class else data_conn_item_class
        CLASSES['exec conn item'] = ExecConnectionItem if not exec_conn_item_class else exec_conn_item_class

        RC_Session.__init__(self, gui=True)

        # flow views
        self.flow_views = {}  # {Script : FlowView}

        # nodes
        Node.complete_default_node_classes()

        # threading
        self.threaded = threaded
        self.threading_bridge = None
        self.gui_parent = gui_parent
        self.threading_bridge = SessionThreadingBridge()
        self.threading_bridge.moveToThread(gui_parent.thread() if threaded else self.thread())
        self.build_flow_view_request.connect(self.threading_bridge.init_flow_view)

        # design
        app = QApplication.instance()
        app.setAttribute(Qt.AA_UseHighDpiPixmaps)
        Design.register_fonts()
        self.design = Design()
        if flow_theme_name:
            self.design.set_flow_theme(name=flow_theme_name)
        if performance_mode:
            self.design.set_performance_mode(performance_mode)


    def create_macro(self, title: str = None, create_default_logs=True,
                     config: dict = None,
                     flow_view_size: list = None) -> MacroScript:
        """Creates and returns a new macro script.
        If a config is provided the title parameter will be ignored and the script will not be initialized,
        which you need to do manually after you made sure that the config doesnt contain other macro nodes
        that have not been loaded yet."""

        script = RC_Session.create_macro(self, title=title, create_default_logs=create_default_logs, config=config)

        self.new_script_created.emit(script)

        self._build_flow_view(script, flow_view_size)

        return script

    def create_script(self, title: str = None, create_default_logs=True,
                      config: dict = None,
                      flow_view_size: list = None) -> Script:
        """Creates and returns a new script.
        If a config is provided the title parameter will be ignored."""

        script = RC_Session.create_script(self, title=title, create_default_logs=create_default_logs, config=config)

        self.new_script_created.emit(script)

        self._build_flow_view(script, flow_view_size)

        return script

    def rename_script(self, script: Script, title: str) -> bool:
        """Renames an existing script; emits script_renamed"""

        success = RC_Session.rename_script(self, script=script, title=title)

        if success:
            self.script_renamed.emit(script)

        return success

    def delete_script(self, script: Script):
        """Deletes an existing script; emits script_deleted"""

        RC_Session.delete_script(self, script=script)
        self.script_deleted.emit(script)


    def _build_flow_view(self, script, view_size):
        """Builds the flow view in the GUI thread"""

        script_config = script.init_config

        flow_view_config = None
        if script_config is not None:
            if 'flow view' not in script_config:  # backwards compatibility
                flow_view_config = script_config['flow']
            else:
                flow_view_config = script_config['flow view']

        flow_view_params = (
            self,
            script,
            script.flow,
            flow_view_config,
            view_size,
            self.gui_parent,
        )

        # build flow view in GUI thread
        self.tmp_data = None
        self.build_flow_view_request.emit(self, flow_view_params)
        while self.tmp_data is None:
            time.sleep(0.001)
        self.tmp_data: FlowView
        flow_view = self.tmp_data

        self.flow_views[script] = flow_view
        self.flow_view_created.emit(script, flow_view)

        return flow_view

    def serialize(self) -> dict:
        """Returns the project as JSON compatible dict to be saved and loaded again using load()"""

        # adds the frontend related data to the abstract project data by the ryvencore Session

        data = RC_Session.serialize(self)

        # MACRO SCRIPTS
        complete_macro_scripts_data = []
        for ms_cfg in data['macro scripts']:
            title = ms_cfg['name']  # script titles are unique!
            macro_script = self._script_from_title(title)
            view = self.flow_views[macro_script]

            # complete script config in FlowView in GUI thread
            self.complete_flow_view_config.connect(view.generate_config_data)
            view._tmp_data = None
            self.complete_flow_view_config.emit(ms_cfg)
            while view._tmp_data is None:
                time.sleep(0.001)
            self.complete_flow_view_config.disconnect(view.generate_config_data)

            complete_macro_scripts_data.append(view._tmp_data)

        # SCRIPTS
        complete_scripts_data = []
        for s_cfg in data['scripts']:
            title = s_cfg['name']  # script title are unique!
            script = self._script_from_title(title)
            view = self.flow_views[script]

            # complete script config in FlowView in GUI thread
            self.complete_flow_view_config.connect(view.generate_config_data)
            view._tmp_data = None
            self.complete_flow_view_config.emit(s_cfg)
            while view._tmp_data is None:
                time.sleep(0.001)
            self.complete_flow_view_config.disconnect(view.generate_config_data)

            complete_scripts_data.append(view._tmp_data)

        complete_data = {
            'macro scripts': complete_macro_scripts_data,
            'scripts': complete_scripts_data,
        }

        return complete_data

    def _script_from_title(self, title: str) -> Script:
        for s in self.scripts:  # all_scripts():
            if s.title == title:
                return s
        return None



# import math
from qtpy.QtCore import QMarginsF
from qtpy.QtCore import QRectF, QPointF, Qt
from qtpy.QtGui import QPainter, QColor, QRadialGradient, QPainterPath, QPen
from qtpy.QtWidgets import QGraphicsItem, QStyleOptionGraphicsItem

from .tools import sqrt
from .tools import pythagoras


class ConnectionItem(QGraphicsItem):
    """The GUI representative for a connection. The classes ExecConnectionItem and DataConnectionItem will be ready
    for reimplementation later, so users can add GUI for the enhancements of DataConnection and ExecConnection,
    like input fields for weights."""

    def __init__(self, connection, session_design):
        super().__init__()

        self.connection = connection

        out = self.connection.out
        inp = self.connection.inp
        out_node = out.node
        inp_node = inp.node
        self.out_item = out_node.port_item(out)
        self.inp_item = inp_node.port_item(inp)

        # self.out = self.connection.out
        # self.out_item = self.connection.out.node.port_item()
        # self.inp = self.connection.inp

        # self.type_ = type_
        self.session_design = session_design
        self.changed = False
        self.path: QPainterPath = None
        self.gradient = None

        # for rendering flow pictures
        self.setCacheMode(QGraphicsItem.DeviceCoordinateCache)

        self.recompute()

    def boundingRect(self):
        if self.path:
            return self.path.boundingRect().marginsAdded(QMarginsF(3, 3, 3, 3))
        else:
            op = self.out_pos()
            ip = self.inp_pos()
            return QRectF(
                0 if op.x() < ip.x() else (ip.x()-op.x()),
                0 if op.y() < ip.y() else (ip.y()-op.y()),
                abs(ip.x()-op.x()),
                abs(ip.y()-op.y())
            )


    def recompute(self):
        """Updated scene position and recomputes the path"""

        self.setPos(self.out_pos())
        self.changed = True
        self.update()
    
    
    def out_pos(self) -> QPointF:
        """The current global scene position of the pin of the output port"""

        return self.out_item.pin.get_scene_center_pos()
    
    def inp_pos(self) -> QPointF:
        """The current global scene position of the pin of the input port"""

        return self.inp_item.pin.get_scene_center_pos()

    @staticmethod
    def dist(p1: QPointF, p2: QPointF) -> float:
        """Returns the diagonal distance between the points using pythagoras"""

        dx = p2.x()-p1.x()
        dy = p2.y()-p1.y()
        return sqrt((dx**2) + (dy**2))


    @staticmethod
    def connection_path(p1: QPointF, p2: QPointF) -> QPainterPath:
        """Returns the painter path for drawing the connection, using the usual cubic connection path by default"""

        return default_cubic_connection_path(p1, p2)



class ExecConnectionItem(ConnectionItem):


    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget=...) -> None:

        theme = self.session_design.flow_theme

        pen = QPen(theme.exec_conn_color, theme.exec_conn_width)
        pen.setStyle(theme.exec_conn_pen_style)
        pen.setCapStyle(Qt.RoundCap)

        c = pen.color()

        # highlight hovered connections
        if self.out_item.pin.hovered or self.inp_item.pin.hovered:
            c = QColor('#c5c5c5')
            pen.setWidth(theme.exec_conn_width*2)

        if self.changed or not self.path:
            self.changed = False

            self.path = self.connection_path(
                QPointF(0, 0),
                self.inp_pos()-self.scenePos()
            )

            w = self.path.boundingRect().width()
            h = self.path.boundingRect().height()
            self.gradient = QRadialGradient(self.path.boundingRect().center(),
                                            pythagoras(w, h) / 2)

            c_r = c.red()
            c_g = c.green()
            c_b = c.blue()

            # this offset will be 1 if inp.x >> out.x and 0 if inp.x < out.x
            # hence, no fade for the gradient if the connection goes backwards
            offset_mult: float = max(
                0,
                min(
                    (self.inp_pos().x() - self.out_pos().x())/200,
                    1
                )
            )

            # and if the input is very far away from the output, decrease the gradient fade so the connection
            # doesn't fully disappear at the ends and stays visible
            if self.inp_pos().x() > self.out_pos().x():
                offset_mult = min(
                    offset_mult,
                    2000/(self.dist(self.inp_pos(), self.out_pos()))
                )
                # zucker.

            self.gradient.setColorAt(0.0, QColor(c_r, c_g, c_b, 255))
            self.gradient.setColorAt(0.75, QColor(c_r, c_g, c_b, 255 - round(55 * offset_mult)))
            self.gradient.setColorAt(0.95, QColor(c_r, c_g, c_b, 255 - round(255 * offset_mult)))

        pen.setBrush(self.gradient)
        painter.setPen(pen)

        painter.drawPath(self.path)


class DataConnectionItem(ConnectionItem):

    def paint(self, painter: QPainter, option: QStyleOptionGraphicsItem, widget=...) -> None:

        theme = self.session_design.flow_theme

        pen = QPen(theme.data_conn_color, theme.data_conn_width)
        pen.setStyle(theme.data_conn_pen_style)
        pen.setCapStyle(Qt.RoundCap)

        c = pen.color()

        # highlight hovered connections
        if self.out_item.pin.hovered or self.inp_item.pin.hovered:
            c = QColor('#c5c5c5')
            pen.setWidth(theme.exec_conn_width*2)


        if self.changed or not self.path:
            self.changed = False

            self.path = self.connection_path(
                QPointF(0, 0),
                self.inp_pos() - self.scenePos()
            )

            w = self.path.boundingRect().width()
            h = self.path.boundingRect().height()
            self.gradient = QRadialGradient(self.path.boundingRect().center(),
                                            pythagoras(w, h) / 2)

            c_r = c.red()
            c_g = c.green()
            c_b = c.blue()

            # this offset will be 1 if inp.x >> out.x and 0 if inp.x < out.x
            # hence, no fade for the gradient if the connection goes backwards
            offset_mult: float = max(
                0,
                min(
                    (self.inp_pos().x() - self.out_pos().x())/200,
                    1
                )
            )

            # and if the input is very far away from the output, decrease the gradient fade so the connection
            # doesn't fully disappear at the ends and stays visible
            if self.inp_pos().x() > self.out_pos().x():
                offset_mult = min(
                    offset_mult,
                    2000/(self.dist(self.inp_pos(), self.out_pos()))
                )
                # zucker.

            self.gradient.setColorAt(0.0, QColor(c_r, c_g, c_b, 255))
            self.gradient.setColorAt(0.75, QColor(c_r, c_g, c_b, 255 - round(55 * offset_mult)))
            self.gradient.setColorAt(0.95, QColor(c_r, c_g, c_b, 255 - round(255 * offset_mult)))

        pen.setBrush(self.gradient)
        painter.setPen(pen)
        painter.drawPath(self.path)


def default_cubic_connection_path(p1: QPointF, p2: QPointF):
    """Returns the nice looking QPainterPath from p1 to p2"""

    path = QPainterPath()

    path.moveTo(p1)

    dx = p2.x() - p1.x()
    adx = abs(dx)
    dy = p2.y() - p1.y()
    ady = abs(dy)
    distance = sqrt((dx ** 2) + (dy ** 2))
    x1, y1 = p1.x(), p1.y()
    x2, y2 = p2.x(), p2.y()

    if ((x1 < x2 - 30) or distance < 100) and (x1 < x2):
        # STANDARD FORWARD
        path.cubicTo(x1 + ((x2 - x1) / 2), y1,
                     x1 + ((x2 - x1) / 2), y2,
                     x2, y2)
    elif x2 < x1 - 100 and adx > ady * 2:
        # STRONG BACKWARDS
        path.cubicTo(x1 + 100 + (x1 - x2) / 10, y1,
                     x1 + 100 + (x1 - x2) / 10, y1 + (dy / 2),
                     x1 + (dx / 2), y1 + (dy / 2))
        path.cubicTo(x2 - 100 - (x1 - x2) / 10, y2 - (dy / 2),
                     x2 - 100 - (x1 - x2) / 10, y2,
                     x2, y2)
    else:
        # STANDARD BACKWARDS
        path.cubicTo(x1 + 100 + (x1 - x2) / 3, y1,
                     x2 - 100 - (x1 - x2) / 3, y2,
                     x2, y2)
    
    return path
